__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_if_keys_exist_in_dict,
    check_if_string_value_is_valid_video_timestamp, check_if_valid_rgb_tuple,
    check_instance, check_str, check_that_column_exist,
    check_that_hhmmss_start_is_before_end, check_valid_boolean,
    check_valid_dataframe, check_valid_lst,
    check_video_and_data_frm_count_align)
from simba.utils.data import (find_frame_numbers_from_time_stamp,
                              slice_roi_dict_for_video)
from simba.utils.enums import Formats, TagNames
from simba.utils.errors import (FrameRangeError, InvalidVideoFileError,
                                NoSpecifiedOutputError)
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    create_directory, find_core_cnt,
                                    find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_df,
                                    read_frm_of_video, remove_a_folder)
from simba.utils.warnings import ROIWarning

STYLE_WIDTH = "width"
STYLE_HEIGHT = "height"
STYLE_LINE_WIDTH = "line width"
STYLE_FONT_SIZE = "font size"
STYLE_FONT_THICKNESS = "font thickness"
STYLE_CIRCLE_SIZE = "circle size"
STYLE_MAX_LINES = "max lines"
STYLE_BG = 'bg'
STYLE_BG_OPACITY = 'bg_opacity'
COLOR = 'color'
SIZE = 'size'
START_TIME = 'start_time'
END_TIME = 'end_time'
BODY_PART = 'body_part'
AUTO = 'AUTO'
ANIMAL_NAME = 'animal_name'

STYLE_KEYS = [
    STYLE_WIDTH,
    STYLE_HEIGHT,
    STYLE_LINE_WIDTH,
    STYLE_FONT_SIZE,
    STYLE_FONT_THICKNESS,
    STYLE_CIRCLE_SIZE,
    STYLE_MAX_LINES,
    STYLE_BG,
    STYLE_BG_OPACITY
]

def path_plot_mp(data: np.ndarray,
                 line_data: np.array,
                 colors: list,
                 video_setting: bool,
                 frame_setting: bool,
                 video_save_dir: str,
                 video_name: str,
                 frame_folder_dir: str,
                 style_attr: dict,
                 roi: Union[dict, None],
                 animal_names: Union[None, List[str]],
                 fps: Union[int, float],
                 clf_attr: dict):

    batch_id, frm_cnt_rng, frm_id_rng = data[0], data[1], data[2]
    if video_setting:
        video_save_path = os.path.join(video_save_dir, f"{batch_id}.mp4")
        FOURCC = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        video_writer = cv2.VideoWriter(video_save_path, FOURCC, int(fps), (style_attr[STYLE_WIDTH], style_attr[STYLE_HEIGHT]))
    for frm_cnt, frm_id in zip(frm_cnt_rng, frm_id_rng):
        if isinstance(style_attr[STYLE_BG], str):
            bg = read_frm_of_video(video_path=style_attr[STYLE_BG], opacity=style_attr[STYLE_BG_OPACITY], frame_index=frm_id)
        else:
            bg = deepcopy(style_attr[STYLE_BG])
        plot_arrs = [x[:frm_cnt, :] for x in line_data]
        clf_attr_cpy = deepcopy(clf_attr)
        if clf_attr is not None:
            for k, v in clf_attr.items(): clf_attr_cpy[k]["clfs"][frm_id + 1 :] = 0

        img = PlottingMixin.make_path_plot(data=plot_arrs,
                                           colors=colors,
                                           width=style_attr[STYLE_WIDTH],
                                           height=style_attr[STYLE_HEIGHT],
                                           max_lines=style_attr[STYLE_MAX_LINES],
                                           bg_clr=bg,
                                           circle_size=style_attr[STYLE_CIRCLE_SIZE],
                                           font_size=style_attr[STYLE_FONT_SIZE],
                                           font_thickness=style_attr[STYLE_FONT_THICKNESS],
                                           line_width=style_attr[STYLE_LINE_WIDTH],
                                           animal_names=animal_names,
                                           clf_attr=clf_attr_cpy,
                                           save_path=None)
        if roi is not None:
            img = PlottingMixin.roi_dict_onto_img(img=img, roi_dict=roi, show_tags=False, show_center=False)

        if video_setting:
            video_writer.write(img)
        if frame_setting:
            frm_name = os.path.join(frame_folder_dir, f"{frm_id}.png")
            cv2.imwrite(frm_name, np.uint8(img))
        print(f"Path frame created: {frm_id}, Video: {video_name}, Processing core: {batch_id}")
    if video_setting:
        video_writer.release()
    return batch_id


class PathPlotterMulticore(ConfigReader, PlottingMixin):
    """
    Class for creating "path plots" videos and/or images detailing the movement paths of
    individual animals in SimBA. Uses multiprocessing.

    :param str config_path: Path to SimBA project config file in Configparser format
    :param bool frame_setting: If True, individual frames will be created.
    :param bool video_setting: If True, compressed videos will be created.
    :param bool last_frame: If True, png of the last frame will be created.
    :param List[str] files_found: Data paths to create path plots for (e.g., ['project_folder/csv/machine_results/MyVideo.csv'])
    :param dict animal_attr: Animal body-parts to use when creating paths and their colors.
    :param Optional[dict] input_style_attr: Plot sttributes (line thickness, color, etc..). If None, then autocomputed. Max lines will be set to 2s.
    :param Optional[dict] input_clf_attr: Dict reprenting classified behavior locations, their color and size. If None, then no classified behavior locations are shown.
    :param Optional[dict] slicing: If Dict, start time and end time of video slice to create path plot from. E.g., {'start_time': '00:00:01', 'end_time': '00:00:03'}. If None, creates path plot for entire video.
    :param Optional[bool] roi: If True, also plots the ROIs associated with the video. Default False.
    :param int cores: Number of cores to use.

    .. note::
       `Visualization tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/path_plot.png
       :width: 300
       :align: center

    .. image:: _static/img/path_plot_mp.gif
       :width: 500
       :align: center

    .. image:: _static/img/path_plot_1.png
       :width: 500
       :align: center

    :example:
    >>> input_style_attr = {'width': 'As input', 'height': 'As input', 'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': 'White', 'max lines': 100}
    >>> animal_attr = {0: ['Ear_right_1', 'Red']}
    >>> input_clf_attr = {0: ['Attack', 'Black', 'Size: 30'], 1: ['Sniffing', 'Red', 'Size: 30']}
    >>> path_plotter = PathPlotterMulticore(config_path=r'MyConfigPath', frame_setting=False, video_setting=True, style_attr=style_attr, animal_attr=animal_attr, files_found=['project_folder/csv/machine_results/MyVideo.csv'], cores=5, clf_attr=clf_attr, print_animal_names=True)
    >>> path_plotter.run()

    References
    ----------
    .. [1] Battivelli, Dorian, Lucas Boldrini, Mohit Jaiswal, Pradnya Patil, Sofia Torchia, Elizabeth Engelen, Luca Spagnoletti, Sarah Kaspar, and Cornelius T. Gross. “Induction of Territorial Dominance and Subordination Behaviors in Laboratory Mice.” Scientific Reports 14, no. 1 (November 19, 2024): 28655. https://doi.org/10.1038/s41598-024-75545-4.
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_paths: List[Union[str, os.PathLike]],
                 animal_attr: dict,
                 style_attr: Optional[Union[Dict[str, Any], None]] = None,
                 clf_attr: Optional[dict] = None,
                 frame_setting: bool = False,
                 video_setting: bool = False,
                 last_frame: bool = False,
                 print_animal_names: bool = True,
                 slicing: Optional[Dict] = None,
                 core_cnt: int = -1,
                 roi: bool = False):

        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        if (not frame_setting) and (not video_setting) and (not last_frame):
            raise NoSpecifiedOutputError(msg="SIMBA ERROR: Please choice to create path frames and/or video path plots", source=self.__class__.__name__)
        check_valid_lst(data=data_paths, source=self.__class__.__name__, valid_dtypes=(str,), min_len=1)
        _ = [check_file_exist_and_readable(x) for x in data_paths]
        check_file_exist_and_readable(file_path=config_path)
        if style_attr is not None:
            check_if_keys_exist_in_dict(data=style_attr, key=STYLE_KEYS, name=f'{self.__class__.__name__} style_attr')
        check_valid_boolean(value=[frame_setting, video_setting, last_frame,print_animal_names, roi], source=self.__class__.__name__)
        if slicing is not None:
            check_if_keys_exist_in_dict(data=slicing, key=[START_TIME, END_TIME], name=f'{self.__class__.__name__} slicing')
            check_if_string_value_is_valid_video_timestamp(value=slicing[START_TIME], name="Video slicing START TIME")
            check_if_string_value_is_valid_video_timestamp(value=slicing[END_TIME], name="Video slicing END TIME")
            check_that_hhmmss_start_is_before_end(start_time=slicing[START_TIME], end_time=slicing[END_TIME], name="SLICE TIME STAMPS")
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        if roi:
            check_file_exist_and_readable(file_path=self.roi_coordinates_path)
            self.read_roi_data()
        if not video_setting and not frame_setting and not last_frame:
            raise NoSpecifiedOutputError(msg='Video, last frame, and frame are all False', source=self.__class__.__name__)
        self.clf_names = None
        if clf_attr is not None:
            check_instance(source=f'{self.__class__.__name__} clf_attr', instance=clf_attr, accepted_types=(dict,))
            for k, v in clf_attr.items():
                check_if_keys_exist_in_dict(data=v, key=[COLOR, SIZE], name=f'{self.__class__.__name__} clf_attr')
            self.clf_names = list(clf_attr.keys())
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        if not os.path.exists(self.path_plot_dir): os.makedirs(self.path_plot_dir)
        self.video_setting, self.frame_setting, self.style_attr, self.data_paths, self.animal_attr, self.clf_attr, self.last_frame, self.roi = video_setting, frame_setting, style_attr, data_paths, animal_attr, clf_attr, last_frame, roi
        self.print_animal_names, self.slicing = print_animal_names, slicing
        self.bp_data, self.data_cols, self.animal_names, self.colors = self.__get_bp_data()
        self.animal_names = None if not self.print_animal_names else self.animal_names
        self.core_cnt = find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt

    def __get_bp_data(self):
        bp_data, data_cols, animal_names, colors = {}, [], [], [],
        for cnt, (k, v) in enumerate(self.animal_attr.items()):
            check_if_keys_exist_in_dict(data=v, key=[COLOR, BODY_PART], name=f'{self.__class__.__name__} animal_attr')
            check_str(name=f'animal {k} body_part', value=v[BODY_PART], options=self.body_parts_lst)
            check_if_valid_rgb_tuple(data=v[COLOR], raise_error=True)
            name = self.find_animal_name_from_body_part_name(bp_name=v[BODY_PART], bp_dict=self.animal_bp_dict)
            bp_data[cnt] = {ANIMAL_NAME: name, 'x': f'{v[BODY_PART]}_x', 'y': f'{v[BODY_PART]}_y', 'p': f'{v[BODY_PART]}_p', COLOR: v[COLOR]}
            data_cols.extend([f'{v[BODY_PART]}_x', f'{v[BODY_PART]}_y'])
            animal_names.append(name)
            colors.append(v[COLOR])
        return bp_data, data_cols, animal_names, colors


    def __get_styles(self, style_attr):
        video_styles = {}
        optimal_font_size, _, _ = PlottingMixin().get_optimal_font_scales(text='MY LONG TEXT STRING', accepted_px_width=int(self.video_info["Resolution_width"] / 10), accepted_px_height=int(self.video_info["Resolution_width"] / 10))
        optimal_circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(int(self.video_info["Resolution_width"]), int(self.video_info["Resolution_height"])), circle_frame_ratio=100)
        video_styles[STYLE_WIDTH] = int(self.video_info["Resolution_width"].values[0]) if style_attr[STYLE_WIDTH] == None else int(style_attr[STYLE_WIDTH])
        video_styles[STYLE_HEIGHT] = int(self.video_info["Resolution_height"].values[0]) if style_attr[STYLE_HEIGHT] == None else int(style_attr[STYLE_HEIGHT])
        video_styles[STYLE_CIRCLE_SIZE] = optimal_circle_size if style_attr[STYLE_CIRCLE_SIZE] == None else int(style_attr[STYLE_CIRCLE_SIZE])
        video_styles[STYLE_LINE_WIDTH] = 2 if style_attr[STYLE_LINE_WIDTH] == None else int(style_attr[STYLE_LINE_WIDTH])
        video_styles[STYLE_FONT_THICKNESS] = 2 if style_attr[STYLE_FONT_THICKNESS] == None else int(style_attr[STYLE_FONT_THICKNESS])
        video_styles[STYLE_FONT_SIZE] = optimal_font_size if style_attr[STYLE_FONT_SIZE] == None else int(style_attr[STYLE_FONT_SIZE])
        video_styles[STYLE_MAX_LINES] = len(self.data_df) if style_attr[STYLE_MAX_LINES] == None else int(self.fps * int(style_attr[STYLE_MAX_LINES]))
        video_styles[STYLE_BG_OPACITY] = None if style_attr[STYLE_BG_OPACITY] == None else float(style_attr[STYLE_BG_OPACITY])
        if isinstance(self.style_attr[STYLE_BG], tuple):
            video_styles[STYLE_BG] = self.style_attr[STYLE_BG]
        if isinstance(self.style_attr[STYLE_BG], (str, int)):
            video_path = find_video_of_file(video_dir=self.video_dir, filename=self.video_name, raise_error=False, warning=True)
            if video_path is None:
                raise InvalidVideoFileError(msg=f'Could not find video in expected location. If using frame or video as background, make sure the data file {self.video_name} is represented in the {self.video_dir} directory', source=self.__class__.__name__)
            if isinstance(self.style_attr[STYLE_BG], (int,)):
                video_styles[STYLE_BG] = read_frm_of_video(video_path=video_path, frame_index=self.style_attr[STYLE_BG], opacity=video_styles[STYLE_BG_OPACITY])
            else:
                if self.slicing is None:
                    check_video_and_data_frm_count_align(video=video_path, data=self.data_df, name=self.video_name, raise_error=False)
                video_styles[STYLE_BG] = video_path
        return video_styles


    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        print(f"Processing path plots for {len(self.data_paths)} video(s)...")
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, _, self.fps = self.read_video_info(video_name=self.video_name)
            self.in_df = read_df(file_path, self.file_type)
            check_valid_dataframe(df=self.in_df, source=file_path, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.data_cols)
            if self.clf_attr is not None:
                check_valid_dataframe(df=self.in_df, source=file_path, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.clf_names)
            self.save_frm_dir = None
            if self.slicing:
                frm_numbers = find_frame_numbers_from_time_stamp(start_time=self.slicing[START_TIME], end_time=self.slicing[END_TIME], fps=self.fps)
                if len(set(frm_numbers) - set(self.in_df.index)) > 0:
                    raise FrameRangeError(msg=f'The chosen time-period ({self.slicing[START_TIME]} - {self.slicing[END_TIME]}) does not exist in {self.video_name}.', source=self.__class__.__name__)
                self.in_df = self.in_df.loc[frm_numbers]
            else:
                frm_numbers = list(range(0, len(self.in_df)))
            self.data_df = self.in_df[self.data_cols]
            video_styles = self.__get_styles(self.style_attr)
            if self.video_setting:
                self.video_save_path = os.path.join(self.path_plot_dir, f"{self.video_name}.mp4")
                self.writer = cv2.VideoWriter(self.video_save_path, self.fourcc, self.fps, (video_styles[STYLE_WIDTH], video_styles[STYLE_HEIGHT]))
                self.video_temp_dir = os.path.join(self.path_plot_dir, self.video_name)
                create_directory(paths=self.video_temp_dir, overwrite=True)
            if self.frame_setting:
                self.save_frm_dir = os.path.join(self.path_plot_dir, self.video_name)
                create_directory(paths=self.save_frm_dir, overwrite=True)
            video_rois, video_roi_names = None, None
            if self.roi:
                video_rois, video_roi_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=self.video_name)
                if len(video_roi_names) == 0:
                    ROIWarning(msg=f'NO ROI data found for video {self.video_name}. Skipping ROI plotting for this video.')
            if self.clf_attr is not None:
                clf_attr = {}
                for k, v in self.clf_attr.items():
                    check_that_column_exist(df=self.in_df, column_name=k, file_name=file_path)
                    clf_attr[k] = {COLOR: v[COLOR], SIZE: v[SIZE], 'clfs': self.in_df[k].values.astype(np.int8), 'positions': self.data_df[[self.bp_data[0]['x'], self.bp_data[0]['y']]].values}
            else:
                clf_attr = None

            line_data = []
            for k, v in self.bp_data.items():
                line_data.append(self.data_df[[v['x'], v['y']]].values)

            if self.last_frame:
                last_frame_save_path = os.path.join(self.path_plot_dir, f"{self.video_name}_final_frame.png")
                if isinstance(video_styles[STYLE_BG], str):
                    bg = read_frm_of_video(video_path=video_styles[STYLE_BG], opacity=video_styles[STYLE_BG_OPACITY], frame_index=len(self.data_df)-1)
                else:
                    bg = deepcopy(video_styles[STYLE_BG])
                last_frm = PlottingMixin.make_path_plot(data=line_data,
                                                        colors=self.colors,
                                                        width=video_styles[STYLE_WIDTH],
                                                        height=video_styles[STYLE_HEIGHT],
                                                        max_lines=video_styles[STYLE_MAX_LINES],
                                                        bg_clr=bg,
                                                        circle_size=video_styles[STYLE_CIRCLE_SIZE],
                                                        font_size=video_styles[STYLE_FONT_SIZE],
                                                        font_thickness=video_styles[STYLE_FONT_THICKNESS],
                                                        line_width=video_styles[STYLE_LINE_WIDTH],
                                                        animal_names=self.animal_names,
                                                        clf_attr=clf_attr,
                                                        save_path=None)
                if video_rois is not None:
                    last_frm = PlottingMixin.roi_dict_onto_img(img=last_frm, roi_dict=video_rois, show_tags=False, show_center=False)
                cv2.imwrite(filename=last_frame_save_path, img=last_frm)
                stdout_success(msg=f'Last path plot frame saved at {last_frame_save_path}')
            if self.video_setting or self.frame_setting:
                frm_cnt_range = np.arange(1, line_data[0].shape[0])
                frm_cnt_range = np.array_split(frm_cnt_range, self.core_cnt)
                frm_id_range = np.array_split(frm_numbers, self.core_cnt)
                frm_range = [(cnt, x, y) for cnt, (x, y) in enumerate(zip(frm_cnt_range, frm_id_range))]
                print(f"Creating path plots, multiprocessing (chunksize: {self.multiprocess_chunksize}, cores: {self.core_cnt})...")
                with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.maxtasksperchild) as pool:
                    constants = functools.partial(path_plot_mp,
                                                  line_data=line_data,
                                                  colors=self.colors,
                                                  video_setting=self.video_setting,
                                                  video_name=self.video_name,
                                                  frame_setting=self.frame_setting,
                                                  video_save_dir=self.video_temp_dir,
                                                  frame_folder_dir=self.save_frm_dir,
                                                  style_attr=video_styles,
                                                  clf_attr=clf_attr,
                                                  animal_names=self.animal_names,
                                                  fps=self.fps,
                                                  roi=video_rois)
                    for cnt, result in enumerate(pool.imap(constants, frm_range, chunksize=self.multiprocess_chunksize)):
                        print(f"Path batch {result+1}/{self.core_cnt} complete...")
                pool.terminate()
                pool.join()

                if self.video_setting:
                    print(f"Joining {self.video_name} multi-processed video...")
                    print(self.video_temp_dir, self.video_save_path)
                    concatenate_videos_in_folder(in_folder=self.video_temp_dir, save_path=self.video_save_path, remove_splits=False)
                video_timer.stop_timer()
                print(f"Path plot video {self.video_name} complete (elapsed time: {video_timer.elapsed_time_str}s) ...")

        self.timer.stop_timer()
        stdout_success(msg=f"Path plot visualizations for {len(self.data_paths)} video(s) created in {self.path_plot_dir} directory", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)


#
# if __name__ == "__main__":
#     style_attr = {STYLE_WIDTH: None,
#                   STYLE_HEIGHT: None,
#                   STYLE_LINE_WIDTH: 5,
#                   STYLE_FONT_SIZE: 5,
#                   STYLE_FONT_THICKNESS: 2,
#                   STYLE_CIRCLE_SIZE: 5,
#                   STYLE_BG: 'video',
#                   STYLE_BG_OPACITY: 100,
#                   STYLE_MAX_LINES: None}
#
#     animal_attr = {0: {'body_part': 'Snout', 'color': (255, 0, 0)}, 1: {'body_part': 'Front Paw R', 'color': (0, 0, 255)}}  #['Ear_right_1', 'Red'], 1: ['Ear_right_2', 'Green']}
#     #clf_attr = {'Rearing': {'color': (155, 1, 10), 'size': 30}}
#     clf_attr = None
#     test = PathPlotterMulticore(config_path=r"C:\troubleshooting\open_field_below\project_folder\project_config.ini",
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  last_frame=True,
#                                  slicing={'start_time': '00:00:00', 'end_time': '00:00:05'},#{'start_time': '00:00:00', 'end_time': '00:00:05'}, #{'start_time': '00:00:00', 'end_time': '00:00:50'}, #{'start_time': '00:00:00', 'end_time': '00:00:01'},, #{'start_time': '00:00:00', 'end_time': '00:00:01'}, #{'start_time': '00:00:00', 'end_time': '00:00:01'},
#                                  style_attr=style_attr,
#                                  animal_attr=animal_attr,
#                                  clf_attr=clf_attr,
#                                  print_animal_names=False,
#                                  core_cnt=1,
#                                  roi=True,
#                                  data_paths=[r"C:\troubleshooting\open_field_below\project_folder\csv\outlier_corrected_movement_location\raw_clip1.csv"])
#     test.run()
#



# if __name__ == "__main__":
#     style_attr = {STYLE_WIDTH: None,
#                   STYLE_HEIGHT: None,
#                   STYLE_LINE_WIDTH: 5,
#                   STYLE_FONT_SIZE: 5,
#                   STYLE_FONT_THICKNESS: 2,
#                   STYLE_CIRCLE_SIZE: 5,
#                   STYLE_BG: 'video',
#                   STYLE_BG_OPACITY: 100,
#                   STYLE_MAX_LINES: None}
#
#     animal_attr = {0: {'body_part': 'Ear_right', 'color': (255, 0, 0)}, 1: {'body_part': 'Center', 'color': (0, 0, 255)}}  #['Ear_right_1', 'Red'], 1: ['Ear_right_2', 'Green']}
#     clf_attr = {'Rearing': {'color': (155, 1, 10), 'size': 30}}
#     test = PathPlotterMulticore(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini",
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  last_frame=False,
#                                  slicing={'start_time': '00:00:00', 'end_time': '00:00:05'},#{'start_time': '00:00:00', 'end_time': '00:00:05'}, #{'start_time': '00:00:00', 'end_time': '00:00:50'}, #{'start_time': '00:00:00', 'end_time': '00:00:01'},, #{'start_time': '00:00:00', 'end_time': '00:00:01'}, #{'start_time': '00:00:00', 'end_time': '00:00:01'},
#                                  style_attr=style_attr,
#                                  animal_attr=animal_attr,
#                                  clf_attr=clf_attr,
#                                  print_animal_names=False,
#                                  core_cnt=-1,
#                                  roi=True,
#                                  data_paths=[r"C:\troubleshooting\RAT_NOR\project_folder\csv\machine_results\03152021_NOB_IOT_8.csv"])
#     test.run()




# animal_attr = {0: {'bp': 'Ear_right', 'color': (255, 0, 0)}, 1: {'bp': 'Tail_base', 'color': (0, 0, 255)}}  #['Ear_right_1', 'Red'], 1: ['Ear_right_2', 'Green']}
# style_attr = {'width': 'As input',
#               'height': 'As input',
#               'line width': 2,
#               'font size': 0.9,
#               'font thickness': 2,
#               'circle size': 5,
#               'bg color': {'type': 'moving', 'opacity': 50, 'frame_index': 200}, #{'type': 'static', 'opacity': 100, 'frame_index': 200}
#               'max lines': 'entire video'}
# clf_attr = {'Nose to Nose': {'color': (155, 1, 10), 'size': 30}, 'Nose to Tailbase': {'color': (155, 90, 10), 'size': 30}}
# clf_attr=None
#
# path_plotter = PathPlotterMulticore(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini',
#                                     frame_setting=False,
#                                     video_setting=True,
#                                     last_frame=True,
#                                     clf_attr=clf_attr,
#                                     input_style_attr=None,
#                                     animal_attr=animal_attr,
#                                     roi=True,
#                                     files_found=['/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/csv/outlier_corrected_movement_location/2022-06-20_NOB_DOT_4.csv'],
#                                     cores=-1,
#                                     slicing = {'start_time': '00:00:00', 'end_time': '00:00:05'}, # {'start_time': '00:00:00', 'end_time': '00:00:05'}, # , #None,
#                                     print_animal_names=False)
# path_plotter.run()


# style_attr = {'width': 'As input',
#               'height': 'As input',
#               'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': {'type': 'static', 'opacity': 100, 'frame_index': 100}, 'max lines': 'entire video'}
# animal_attr = {0: ['Ear_right_1', 'Red'], 1: ['Ear_right_2', 'Green']}
# # clf_attr = {0: ['Attack', 'Black', 'Size: 30']}
# # # #
# clf_attr = None
# style_attr = None
#
# path_plotter = PathPlotterMulticore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                     frame_setting=False,
#                                     video_setting=True,
#                                     last_frame=True,
#                                     input_clf_attr=clf_attr,
#                                     input_style_attr=style_attr,
#                                     animal_attr=animal_attr,
#                                     files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                     cores=5,
#                                     slicing = None)
# path_plotter.run()

#


# style_attr = None
# style_attr = {'width': 'As input',
#               'height': 'As input',
#               'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': {'opacity': 100}, 'max lines': 10000}
# animal_attr = {0: ['mouth', 'Red']}
# clf_attr = {0: ['Freezing', 'Black', 'Size: 10'], 1: ['Normal Swimming', 'Red', 'Size: 10']}
# #
# # clf_attr = None
#
#
# path_plotter = PathPlotterMulticore(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                                     frame_setting=False,
#                                     video_setting=True,
#                                     last_frame=True,
#                                     input_clf_attr=clf_attr,
#                                     input_style_attr=style_attr,
#                                     animal_attr=animal_attr,
#                                     files_found=['/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/machine_results/SF8.csv'],
#                                     cores=5)
# path_plotter.run()



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from PIL import Image

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_float, check_if_valid_rgb_tuple,
                                check_str, check_that_column_exist,
                                check_valid_boolean,
                                check_video_and_data_frm_count_align)
from simba.utils.data import create_color_palette
from simba.utils.enums import (ConfigKey, Dtypes, Formats, Options, TagNames,
                               TextOptions)
from simba.utils.errors import (InvalidInputError, NoDataError,
                                NoSpecifiedOutputError)
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (find_all_videos_in_project, get_fn_ext,
                                    get_video_meta_data, read_config_entry,
                                    read_df)
from simba.utils.warnings import FrameRangeWarning

CIRCLE_SCALE = 'circle_scale'
FONT_SIZE = 'font_size'
SPACE_SCALE = 'spacing_scale'
TEXT_THICKNESS = 'text_thickness'
TEXT_SETTING_KEYS = ['circle_scale', 'font_size', 'spacing_scale', 'text_thickness']
CENTER_BP_TXT = ['centroid', 'center']

FOURCC = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)

class PlotSklearnResultsSingleCore(ConfigReader, TrainModelMixin, PlottingMixin):
    """
    Plot classification results overlays on videos. Results are stored in the
    `project_folder/frames/output/sklearn_results` directory of the SimBA project.

    .. note::
       For improved run-time, see :meth:`simba.plotting.plot_clf_results_mp.PlotSklearnResultsMultiProcess` for multiprocess class.
       Scikit visualization documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-10-sklearn-visualization__.

    .. image:: _static/img/sklearn_visualization.gif
       :width: 600
       :align: center

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format
    :param Optional[bool] video_setting: If True, SimBA will create compressed videos. Default True.
    :param Optional[bool] frame_setting: If True, SimBA will create individual frames. Default True.
    :param Optional[str] video_file_path: Path to video file to create classification visualizations for. If None, then all the videos in the csv/machine_results will be used. Default None.
    :param Optional[Union[Dict[str, float], bool]] text_settings: Dictionary holding the circle size, font size, spacing size, and text thickness of the printed text. If None, then these are autocomputed.
    :param Optional[bool] rotate: If True, the output video will be rotated 90 degrees from the input. Default False.
    :param Optional[str] palette: The name of the palette used for the pose-estimation key-points. Default ``Set1``.
    :param Optional[bool] print_timers: If True, the output video will have the cumulative time of the classified behaviours overlaid. Default True.

    :example:
    >>> text_settings = {'circle_scale': 5, 'font_size': 5, 'spacing_scale': 2, 'text_thickness': 10}
    >>> test = PlotSklearnResultsSingleCore(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
    >>>                                       video_setting=True,
    >>>                                       frame_setting=False,
    >>>                                       video_file_path='Together_1.avi',
    >>>                                       print_timers=True,
    >>>                                       text_settings=text_settings,
    >>>                                       rotate=False)
    >>> test.run()

    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 video_setting: bool = True,
                 frame_setting: bool = False,
                 video_paths: Optional[Union[List[Union[str, os.PathLike]], Union[str, os.PathLike]]] = None,
                 rotate: bool = False,
                 animal_names: bool = False,
                 show_pose: bool = True,
                 font_size: Optional[Union[int, float]] = None,
                 space_size: Optional[Union[int, float]] = None,
                 text_opacity: Optional[Union[int, float]] = None,
                 text_thickness: Optional[Union[int, float]] = None,
                 circle_size: Optional[Union[int, float]] = None,
                 pose_palette: Optional[str] = 'Set1',
                 print_timers: bool = True,
                 text_clr: Tuple[int, int,int] = (255, 255, 255),
                 text_bg_clr: Tuple[int, int,int] = (0, 0, 0)):

        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        PlottingMixin.__init__(self)
        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        for i in [video_setting, frame_setting, rotate, print_timers, animal_names, show_pose]:
            check_valid_boolean(value=i, source=self.__class__.__name__, raise_error=True)
        if (not video_setting) and (not frame_setting):
            raise NoSpecifiedOutputError(msg="Please choose to create a video and/or frames. SimBA found that you ticked neither video and/or frames", source=self.__class__.__name__)
        if font_size is not None: check_float(name=f'{self.__class__.__name__} font_size', value=font_size, min_value=0.1)
        if space_size is not None: check_float(name=f'{self.__class__.__name__} space_size', value=space_size, min_value=0.1)
        if text_thickness is not None: check_float(name=f'{self.__class__.__name__} text_thickness', value=text_thickness, min_value=0.1)
        if circle_size is not None: check_float(name=f'{self.__class__.__name__} text_thickness', value=circle_size, min_value=0.1)
        if circle_size is not None: check_float(name=f'{self.__class__.__name__} text_thickness', value=circle_size, min_value=0.1)
        if text_opacity is not None: check_float(name=f'{self.__class__.__name__} text_opacity', value=text_opacity, min_value=0.1)
        check_if_valid_rgb_tuple(data=text_bg_clr, source=f'{self.__class__.__name__} text_bg_clr')
        check_if_valid_rgb_tuple(data=text_clr, source=f'{self.__class__.__name__} text_clr')
        self.text_color, self.text_bg_color = text_clr, text_bg_clr
        self.video_paths, self.print_timers = video_paths, print_timers
        if self.video_paths is None:
            self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
            if len(self.video_paths) == 0:
                raise NoDataError(msg=f'Cannot create classification videos. No videos exist in {self.video_dir} directory', source=self.__class__.__name__)
        self.video_setting, self.frame_setting, self.rotate, self.text_opacity = video_setting, frame_setting, rotate, text_opacity
        self.circle_size, self.font_size, self.animal_names = circle_size, font_size, animal_names
        self.text_thickness, self.space_size, self.show_pose = text_thickness, space_size, show_pose
        self.pose_threshold = read_config_entry(self.config, ConfigKey.THRESHOLD_SETTINGS.value, ConfigKey.SKLEARN_BP_PROB_THRESH.value, Dtypes.FLOAT.value, 0.00)
        if not os.path.exists(self.sklearn_plot_dir):
            os.makedirs(self.sklearn_plot_dir)
        pose_palettes = Options.PALETTE_OPTIONS_CATEGORICAL.value + Options.PALETTE_OPTIONS.value
        check_str(name=f'{self.__class__.__name__} pose_palette', value=pose_palette, options=pose_palettes)
        self.clr_lst = create_color_palette(pallete_name=pose_palette, increments=len(self.body_parts_lst)+1)
        if isinstance(self.video_paths, str): self.video_paths = [video_paths]
        elif isinstance(self.video_paths, list): self.video_paths = video_paths
        else:
            raise InvalidInputError(msg=f'video_paths has to be a path of a list of paths. Got {type(video_paths)}', source=self.__class__.__name__)
        for video_path in self.video_paths:
            video_name = get_fn_ext(filepath=video_path)[1]
            data_path = os.path.join(self.machine_results_dir, f'{video_name}.{self.file_type}')
            if not os.path.isfile(data_path): raise NoDataError(msg=f'Cannot create classification videos for {video_name}. Expected classification data at location {data_path} but file does not exist', source=self.__class__.__name__)


    def __get_print_settings(self):
        self.video_text_thickness = TextOptions.TEXT_THICKNESS.value if self.text_thickness is None else int(max(self.text_thickness, 1))
        longest_str = str(max(['TIMERS:', 'ENSEMBLE PREDICTION:'] + self.clf_names, key=len))
        optimal_font_size, _, optimal_spacing_scale = self.get_optimal_font_scales(text=longest_str, accepted_px_width=int(self.video_meta_data["width"] / 3), accepted_px_height=int(self.video_meta_data["height"] / 10), text_thickness=self.video_text_thickness)
        optimal_circle_size = self.get_optimal_circle_size(frame_size=(self.video_meta_data["width"], self.video_meta_data["height"]), circle_frame_ratio=100)
        self.video_circle_size = optimal_circle_size if self.circle_size is None else int(self.circle_size)
        self.video_font_size = optimal_font_size if self.font_size is None else self.font_size
        self.video_space_size = optimal_spacing_scale if self.space_size is None else int(max(self.space_size, 1))
        self.video_text_opacity = 0.8 if self.text_opacity is None else float(self.text_opacity)

    def run(self):
        for video_cnt, video_path in enumerate(self.video_paths):
            _, self.video_name, _ = get_fn_ext(video_path)
            self.data_path = os.path.join(self.machine_results_dir, f'{self.video_name}.{self.file_type}')
            self.data_df = read_df(self.data_path, self.file_type).reset_index(drop=True).fillna(0)
            if self.show_pose: check_that_column_exist(df=self.data_df, column_name=self.bp_col_names, file_name=self.data_path)
            self.video_meta_data = get_video_meta_data(video_path=video_path)
            height, width = deepcopy(self.video_meta_data["height"]), deepcopy(self.video_meta_data["width"])
            self.save_path = os.path.join(self.sklearn_plot_dir, f"{self.video_name}.mp4")
            if self.frame_setting:
                self.video_frame_dir = os.path.join(self.sklearn_plot_dir, self.video_name)
                if not os.path.exists(self.video_frame_dir): os.makedirs(self.video_frame_dir)
            if self.rotate:
                self.video_meta_data["height"], self.video_meta_data["width"] = (width, height)
            self.clf_timers = {k: 0 for k in self.clf_names}
            check_video_and_data_frm_count_align(video=video_path, data=self.data_df, name=self.video_name, raise_error=False)
            check_that_column_exist(df=self.data_df, column_name=list(self.clf_timers.keys()), file_name=self.data_path)
            self.writer = cv2.VideoWriter(self.save_path, FOURCC, self.video_meta_data['fps'], (self.video_meta_data["width"], self.video_meta_data["height"]))
            self.__get_print_settings()
            self.cap = cv2.VideoCapture(video_path)

            frm_idx = 0
            while self.cap.isOpened():
                ret, self.frame = self.cap.read()
                if ret:
                    clr_cnt = 0
                    for animal_name, animal_data in self.animal_bp_dict.items():
                        if self.show_pose:
                            for bp_num in range(len(animal_data["X_bps"])):
                                x_bp, y_bp, p_bp = (animal_data["X_bps"][bp_num], animal_data["Y_bps"][bp_num], animal_data["P_bps"][bp_num])
                                bp_cords = self.data_df.loc[frm_idx, [x_bp, y_bp, p_bp]]
                                if bp_cords[p_bp] >= self.pose_threshold:
                                    self.frame = cv2.circle(self.frame, (int(bp_cords[x_bp]), int(bp_cords[y_bp])), self.video_circle_size, self.clr_lst[clr_cnt], -1)
                                clr_cnt+=1
                        if self.animal_names:
                            x_bp, y_bp, p_bp = (animal_data["X_bps"][0], animal_data["Y_bps"][0], animal_data["P_bps"][0])
                            bp_cords = self.data_df.loc[frm_idx, [x_bp, y_bp, p_bp]]
                            cv2.putText(self.frame, animal_name, (int(bp_cords[x_bp]), int(bp_cords[y_bp])), self.font, self.video_font_size, self.clr_lst[0], self.video_text_thickness)
                    if self.rotate:
                        self.frame = np.array(Image.fromarray(self.frame).rotate(90, Image.BICUBIC, expand=True))
                    if self.print_timers:
                        self.frame = PlottingMixin().put_text(img=self.frame, text="TIMERS:", pos=(TextOptions.BORDER_BUFFER_Y.value, ((self.video_meta_data["height"] - self.video_meta_data["height"]) + self.video_space_size)), font_size=self.video_font_size, font_thickness=self.video_text_thickness, font=self.font, text_bg_alpha=self.video_text_opacity, text_color_bg=self.text_bg_color, text_color=self.text_color)
                    self.add_spacer = 2
                    for clf_name, clf_time in self.clf_timers.items():
                        frame_results = self.data_df.loc[frm_idx, clf_name]
                        self.clf_timers[clf_name] += (frame_results / self.video_meta_data['fps'])
                        if self.print_timers:
                            self.frame = PlottingMixin().put_text(img=self.frame, text=f"{clf_name} {round(self.clf_timers[clf_name], 2)}", pos=(TextOptions.BORDER_BUFFER_Y.value, ((self.video_meta_data["height"] - self.video_meta_data["height"]) + self.video_space_size * self.add_spacer)), font_size=self.video_font_size, font_thickness=self.video_text_thickness, font=self.font, text_bg_alpha=self.video_text_opacity, text_color_bg=self.text_bg_color, text_color=self.text_color)
                            self.add_spacer += 1
                    self.frame = PlottingMixin().put_text(img=self.frame, text="ENSEMBLE PREDICTION:", pos=(TextOptions.BORDER_BUFFER_Y.value, ((self.video_meta_data["height"] - self.video_meta_data["height"]) + self.video_space_size * self.add_spacer)), font_size=self.video_font_size, font_thickness=self.video_text_thickness, font=self.font, text_bg_alpha=self.video_text_opacity, text_color_bg=self.text_bg_color, text_color=self.text_color)
                    self.add_spacer += 1
                    for clf_name, clf_time in self.clf_timers.items():
                        if self.data_df.loc[frm_idx, clf_name] == 1:
                            self.frame = PlottingMixin().put_text(img=self.frame, text=clf_name, pos=(TextOptions.BORDER_BUFFER_Y.value, (self.video_meta_data["height"] - self.video_meta_data["height"]) + self.video_space_size * self.add_spacer), font_size=self.video_font_size, font_thickness=self.video_text_thickness, font=self.font, text_color=TextOptions.COLOR.value, text_bg_alpha=self.video_text_opacity)
                            self.add_spacer += 1
                    if self.video_setting:
                        self.writer.write(self.frame.astype(np.uint8))
                    if self.frame_setting:
                        frame_save_name = os.path.join(self.video_frame_dir, f"{frm_idx}.png")
                        cv2.imwrite(frame_save_name, self.frame)
                    frm_idx += 1
                    print(f'Frame: {frm_idx} / {self.video_meta_data["frame_count"]}. Video: {self.video_name} ({video_cnt + 1}/{len(self.video_paths)})')
                else:
                    FrameRangeWarning(msg=f'Could not read frame {frm_idx} in video {video_path}. Stoping video creation.')
                    break
            print(f"Video {self.video_name} saved at {self.save_path}...")
            self.cap.release()
            self.writer.release()

        self.timer.stop_timer()
        stdout_success(msg=f"{len(self.video_paths)} visualization(s) created in {self.sklearn_plot_dir} directory", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)

# test = PlotSklearnResultsSingleCore(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini",
#                                     video_setting=True,
#                                     frame_setting=False,
#                                     video_paths=r"C:\troubleshooting\RAT_NOR\project_folder\videos\03152021_NOB_IOT_8.mp4",
#                                     print_timers=True,
#                                     rotate=True,
#                                     animal_names=True)
# test.run()



import argparse
import os
from itertools import product
from typing import Union

import numpy as np
import pandas as pd
from numba.typed import List

from simba.feature_extractors.perimeter_jit import jitted_hull
from simba.mixins.abstract_classes import AbstractFeatureExtraction
from simba.mixins.circular_statistics import CircularStatisticsMixin
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.statistics_mixin import Statistics
from simba.mixins.timeseries_features_mixin import TimeseriesFeatureMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_if_filepath_list_is_empty)
from simba.utils.read_write import (SimbaTimer, get_fn_ext, read_df,
                                    stdout_success, write_df)

NOSE = 'nose'
LEFT_SIDE = 'left_side'
RIGHT_SIDE = 'right_side'
LEFT_EAR = 'left_ear'
RIGHT_EAR = 'right_ear'
CENTER = 'center'
TAIL_BASE = 'tail_base'
TAIL_CENTER = 'tail_center'
TAIL_TIP = 'tail_tip'

TIME_WINDOWS = np.array([0.25, 0.5, 1.0, 2.0])

class MitraFeatureExtractor(ConfigReader,
                            AbstractFeatureExtraction):


    """
    Feature extractor used for grooming and rearing classifications. A subset is of created features are further used for straub-tail classifications.

    :param Union[str, os.PathLike] config_path: Path to SimBA project_config.ini.
    :return: None. Featurized pose-estimation data is saved in the simba project `project_folder/csv/features_extracted` directory.

    :example:
    >>> feature_extractor = MitraFeatureExtractor(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini")
    >>> feature_extractor.run()

    References
    ----------
    .. [1] Lazaro et al., Brainwide Genetic Capture for Conscious State Transitions, `biorxiv`, doi: https://doi.org/10.1101/2025.03.28.646066
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=True, create_logger=False)
        check_if_filepath_list_is_empty(filepaths=self.outlier_corrected_paths, error_msg=f'No data files found in {self.outlier_corrected_dir} directory.')
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.outlier_corrected_paths)

    def run(self):
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            df = read_df(file_path=file_path, file_type=self.file_type)
            results = pd.DataFrame()
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(filepath=file_path)
            save_path = os.path.join(self.features_dir, video_name + f'.{self.file_type}')
            print(f'Featurizing video {video_name} ...(Video {file_cnt+1}/{len(self.outlier_corrected_paths)})')
            _, px_per_mm, fps = self.read_video_info(video_name=video_name)
            shifted_ =  df.shift(periods=1).combine_first(df)
            nose_arr = df[[f'{NOSE}_x', f'{NOSE}_y']].values.astype(np.float32)
            p_arr = df[self.animal_bp_dict['Animal_1']['P_bps']].values.astype(np.float32)
            tailbase_arr = df[[f'{TAIL_BASE}_x', f'{TAIL_BASE}_y']].values.astype(np.float32)
            left_ear_arr = df[[f'{LEFT_EAR}_x', f'{LEFT_EAR}_y']].values.astype(np.float32)
            right_ear_arr = df[[f'{RIGHT_EAR}_x', f'{RIGHT_EAR}_y']].values.astype(np.float32)
            center_arr = df[[f'{CENTER}_x', f'{CENTER}_y']].values.astype(np.float32)
            lat_left_arr = df[[f'{LEFT_SIDE}_x', f'{LEFT_SIDE}_y']].values.astype(np.float32)
            lat_right_arr = df[[f'{RIGHT_SIDE}_x', f'{RIGHT_SIDE}_y']].values.astype(np.float32)
            tail_center_arr = df[[f'{TAIL_CENTER}_x', f'{TAIL_CENTER}_y']].values.astype(np.float32)
            tail_tip_arr = df[[f'{TAIL_TIP}_x', f'{TAIL_TIP}_y']].values.astype(np.float32)
            animal_hull_arr = df[[f'{LEFT_EAR}_x', f'{LEFT_EAR}_y', f'{RIGHT_EAR}_x', f'{RIGHT_EAR}_y', f'{NOSE}_x', f'{NOSE}_y', f'{LEFT_SIDE}_x', f'{LEFT_SIDE}_y', f'{RIGHT_SIDE}_x', f'{RIGHT_SIDE}_y', f'{TAIL_BASE}_x', f'{TAIL_BASE}_y']].values.astype(np.float32).reshape(len(df), 6, 2)
            animal_head_arr = df[[f'{LEFT_EAR}_x', f'{LEFT_EAR}_y', f'{RIGHT_EAR}_x', f'{RIGHT_EAR}_y', f'{NOSE}_x', f'{NOSE}_y']].values.astype(np.float32).reshape(len(df), 3, 2)
            animal_body_arr = df[[f'{LEFT_EAR}_x', f'{LEFT_EAR}_y', f'{RIGHT_EAR}_x', f'{RIGHT_EAR}_y', f'{LEFT_SIDE}_x', f'{LEFT_SIDE}_y', f'{RIGHT_SIDE}_x', f'{RIGHT_SIDE}_y', f'{TAIL_BASE}_x', f'{TAIL_BASE}_y']].values.astype(np.float32).reshape(len(df), 5, 2)
            animal_lower_body_arr = df[[f'{LEFT_SIDE}_x', f'{LEFT_SIDE}_y', f'{RIGHT_SIDE}_x', f'{RIGHT_SIDE}_y', f'{TAIL_BASE}_x', f'{TAIL_BASE}_y']].values.astype(np.float32).reshape(len(df), 3, 2)
            animal_upper_body_arr = df[[f'{LEFT_EAR}_x', f'{LEFT_EAR}_y', f'{RIGHT_EAR}_x', f'{RIGHT_EAR}_y', f'{NOSE}_x', f'{NOSE}_y', f'{LEFT_SIDE}_x', f'{LEFT_SIDE}_y', f'{RIGHT_SIDE}_x', f'{RIGHT_SIDE}_y']].values.astype(np.float32).reshape(len(df), 5, 2)
            left_body_arr = df[[f'{LEFT_EAR}_x', f'{LEFT_EAR}_y', f'{NOSE}_x', f'{NOSE}_y', f'{LEFT_SIDE}_x', f'{LEFT_SIDE}_y', f'{TAIL_BASE}_x', f'{TAIL_BASE}_y', f'{CENTER}_x', f'{CENTER}_y']].values.astype(np.float32).reshape(len(df), 5, 2)
            right_body_arr = df[[f'{RIGHT_EAR}_x', f'{RIGHT_EAR}_y', f'{NOSE}_x', f'{NOSE}_y', f'{RIGHT_SIDE}_x', f'{RIGHT_SIDE}_y', f'{TAIL_BASE}_x', f'{TAIL_BASE}_y', f'{CENTER}_x', f'{CENTER}_y']].values.astype(np.float32).reshape(len(df), 5, 2)
            direction_degrees = CircularStatisticsMixin().direction_three_bps(nose_loc=nose_arr, left_ear_loc=left_ear_arr, right_ear_loc=right_ear_arr).astype(np.float32)

            # GEOMETRY FEATURES
            print('Compute geometry features...')
            results['GEOMETRY_FRAME_HULL_LENGTH'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=nose_arr, location_2=tailbase_arr, px_per_mm=px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_HULL_WIDTH'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=lat_left_arr, location_2=lat_right_arr, px_per_mm=px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_HULL_AREA'] = (jitted_hull(points=animal_hull_arr, target='area') / px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_BODY_AREA'] = (jitted_hull(points=animal_body_arr, target='area') / px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_LOWER_BODY_AREA'] = (jitted_hull(points=animal_lower_body_arr, target='area') / px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_UPPER_BODY_AREA'] = (jitted_hull(points=animal_upper_body_arr, target='area') / px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_HEAD_AREA'] = (jitted_hull(points=animal_head_arr, target='area') / px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_LEFT_BODY_AREA'] = (jitted_hull(points=left_body_arr, target='area') / px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_RIGHT_BODY_AREA'] = (jitted_hull(points=right_body_arr, target='area') / px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_TAIL_LENGTH'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=tailbase_arr, location_2=tail_tip_arr, px_per_mm=px_per_mm).astype(np.int32)
            results['GEOMETRY_FRAME_EAR_DISTANCE'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=left_ear_arr, location_2=right_ear_arr, px_per_mm=px_per_mm).astype(np.int32)

            for time, feature in product(TIME_WINDOWS, ['HULL_LENGTH', 'HULL_WIDTH', 'HULL_AREA', 'BODY_AREA', 'LOWER_BODY_AREA', 'UPPER_BODY_AREA', 'HEAD_AREA', 'LEFT_BODY_AREA', 'RIGHT_BODY_AREA', 'TAIL_LENGTH', 'EAR_DISTANCE']):
                results[f'GEOMETRY_MEAN_{feature}_{time}'] = results[f'GEOMETRY_FRAME_{feature}'].rolling(int(time * fps), min_periods=1).mean().fillna(0).astype(np.int32)
                results[f'GEOMETRY_VAR_{feature}_{time}'] = results[f'GEOMETRY_FRAME_{feature}'].rolling(int(time * fps), min_periods=1).var().fillna(0).astype(np.float32)
                results[f'GEOMETRY_SUM_{feature}_{time}'] = results[f'GEOMETRY_FRAME_{feature}'].rolling(int(time * fps), min_periods=1).sum().fillna(0).astype(np.int32)

            for feature in ['HULL_LENGTH', 'HULL_WIDTH', 'HULL_AREA', 'BODY_AREA', 'LOWER_BODY_AREA', 'UPPER_BODY_AREA', 'HEAD_AREA', 'LEFT_BODY_AREA', 'RIGHT_BODY_AREA', 'TAIL_LENGTH', 'EAR_DISTANCE']:
                sliding_skew = pd.DataFrame(Statistics.sliding_z_scores(data=results[f'GEOMETRY_FRAME_{feature}'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=int(fps)), columns=[f'GEOMETRY_{feature}_SLIDING_Z_SCORE_250', f'GEOMETRY_{feature}_SLIDING_Z_SCORE_500', f'GEOMETRY_{feature}_SLIDING_Z_SCORE_1000', f'GEOMETRY_{feature}_SLIDING_Z_SCORE_2000'])
                sliding_mad_median = pd.DataFrame(Statistics.sliding_mad_median_rule(data=results[f'GEOMETRY_FRAME_{feature}'].values.astype(np.float32), k=0.5, time_windows=TIME_WINDOWS, fps=fps), columns=[f'GEOMETRY_{feature}_SLIDING_MAD_MEDIAN_0.5_250', f'GEOMETRY_{feature}_SLIDING_MAD_MEDIAN_0.5_500', f'GEOMETRY_{feature}_SLIDING_MAD_MEDIAN_0.5_1000', f'GEOMETRY_{feature}_SLIDING_MAD_MEDIAN_0.5_2000'])
                results = pd.concat([results, sliding_skew, sliding_mad_median], axis=1)

            for feature in ['HULL_LENGTH', 'HULL_WIDTH', 'HULL_AREA', 'BODY_AREA', 'LOWER_BODY_AREA', 'UPPER_BODY_AREA', 'HEAD_AREA', 'LEFT_BODY_AREA', 'RIGHT_BODY_AREA', 'TAIL_LENGTH', 'EAR_DISTANCE']:
                statistics = List(['mac', 'rms'])
                x = TimeseriesFeatureMixin.sliding_descriptive_statistics(data=results[f'GEOMETRY_FRAME_{feature}'].values.astype(np.float32), window_sizes=TIME_WINDOWS, sample_rate=int(fps), statistics=statistics)
                for i in range(x.shape[0]):
                    v = pd.DataFrame(x[i], columns=[f'GEOMETRY_{feature}_{statistics[i]}_250', f'GEOMETRY_{feature}_{statistics[i]}_500', f'GEOMETRY_{feature}_{statistics[i]}_1000', f'GEOMETRY_{feature}_{statistics[i]}_2000'])
                    results = pd.concat([results, v], axis=1)

            upper_lower_body_size_correlations = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['GEOMETRY_FRAME_UPPER_BODY_AREA'].values.astype(np.float32), sample_2=results['GEOMETRY_FRAME_LOWER_BODY_AREA'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['GEOMETRY_UPPER_LOWER_BODY_SIZE_SPEARMAN_250', 'GEOMETRY_UPPER_LOWER_BODY_SIZE_SPEARMAN_500', 'GEOMETRY_UPPER_LOWER_BODY_SIZE_SPEARMAN_1000', 'GEOMETRY_UPPER_LOWER_BODY_SIZE_SPEARMAN_2000']).astype(np.float32)
            hull_head_correlations = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['GEOMETRY_FRAME_HULL_AREA'].values.astype(np.float32), sample_2=results['GEOMETRY_FRAME_HEAD_AREA'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['GEOMETRY_HULL_HEAD_SIZE_SPEARMAN_250', 'GEOMETRY_HULL_HEAD_BODY_SIZE_SPEARMAN_500', 'GEOMETRY_HULL_HEAD_SIZE_SPEARMAN_1000', 'GEOMETRY_HULL_HEAD_SIZE_SPEARMAN_2000']).astype(np.float32)
            hull_tail_length_correlations = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['GEOMETRY_FRAME_HULL_LENGTH'].values.astype(np.float32), sample_2=results['GEOMETRY_FRAME_TAIL_LENGTH'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['GEOMETRY_HULL_TAIL_LENGTH_SPEARMAN_250', 'GEOMETRY_HULL_TAIL_LENGTH_SPEARMAN_500', 'GEOMETRY_HULL_TAIL_LENGTH_SPEARMAN_1000', 'GEOMETRY_HULL_TAIL_LENGTH_SPEARMAN_2000']).astype(np.float32)
            left_body_right_body_correlations = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['GEOMETRY_FRAME_LEFT_BODY_AREA'].values.astype(np.float32), sample_2=results['GEOMETRY_FRAME_RIGHT_BODY_AREA'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['GEOMETRY_LEFT_RIGHT_BODY_SPEARMAN_250', 'GEOMETRY_LEFT_RIGHT_BODY_SPEARMAN_500', 'GEOMETRY_LEFT_RIGHT_BODY_SPEARMAN_1000', 'GEOMETRY_LEFT_RIGHT_BODY_SPEARMAN_2000']).astype(np.float32)
            results = pd.concat([results, upper_lower_body_size_correlations, hull_head_correlations, hull_tail_length_correlations, left_body_right_body_correlations], axis=1)

            # CIRCULAR FEATURES
            print('Compute circular features...')
            results['CIRCULAR_FRAME_HULL_3POINT_ANGLE'] = FeatureExtractionMixin.angle3pt_serialized(data=np.hstack([nose_arr, center_arr, tailbase_arr]))
            results['CIRCULAR_FRAME_TAIL_3POINT_ANGLE'] = FeatureExtractionMixin.angle3pt_serialized(data=np.hstack([tailbase_arr, tail_center_arr, tail_tip_arr]))
            results['CIRCULAR_FRAME_HEAD_3POINT_ANGLE'] = FeatureExtractionMixin.angle3pt_serialized(data=np.hstack([left_ear_arr, nose_arr, right_ear_arr]))
            results['CIRCULAR_INSTANTANEOUS_ANGULAR_VELOCITY'] = CircularStatisticsMixin.instantaneous_angular_velocity(data=direction_degrees, bin_size=1)
            angular_difference = pd.DataFrame(CircularStatisticsMixin.sliding_angular_diff(data=direction_degrees, time_windows=TIME_WINDOWS, fps=int(fps)), columns=['CIRCULAR_HEAD_DIRECTION_ANGULAR_DIFFERENCE_250', 'CIRCULAR_HEAD_DIRECTION_ANGULAR_DIFFERENCE_500', 'CIRCULAR_HEAD_DIRECTION_ANGULAR_DIFFERENCE_1000',  'CIRCULAR_HEAD_DIRECTION_ANGULAR_DIFFERENCE_2000'])
            rao_spacing = pd.DataFrame(CircularStatisticsMixin.sliding_rao_spacing(data=direction_degrees, time_windows=TIME_WINDOWS, fps=int(fps)), columns=['CIRCULAR_HEAD_DIRECTION_RAO_SPACING_250', 'CIRCULAR_HEAD_DIRECTION_RAO_SPACING_500', 'CIRCULAR_HEAD_DIRECTION_RAO_SPACING_1000', 'CIRCULAR_HEAD_DIRECTION_RAO_SPACING_2000'])
            circular_range = pd.DataFrame(CircularStatisticsMixin.sliding_circular_range(data=direction_degrees, time_windows=TIME_WINDOWS, fps=int(fps)), columns=['CIRCULAR_HEAD_DIRECTION_RANGE_250', 'CIRCULAR_HEAD_DIRECTION_RANGE_500', 'CIRCULAR_HEAD_DIRECTION_RANGE_1000', 'CIRCULAR_HEAD_DIRECTION_RANGE_2000'])
            circular_std = pd.DataFrame(CircularStatisticsMixin.sliding_circular_std(data=direction_degrees, time_windows=TIME_WINDOWS, fps=int(fps)), columns=['CIRCULAR_HEAD_DIRECTION_STD_250', 'CIRCULAR_HEAD_DIRECTION_STD_500', 'CIRCULAR_HEAD_DIRECTION_STD_1000', 'CIRCULAR_HEAD_DIRECTION_STD_2000'])
            head_hull_angular_corr = pd.DataFrame(CircularStatisticsMixin.sliding_circular_correlation(sample_1=results['CIRCULAR_FRAME_HULL_3POINT_ANGLE'].values.astype(np.float32), sample_2=results['CIRCULAR_FRAME_HEAD_3POINT_ANGLE'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['CIRCULAR_HULL_HEAD_3POINT_ANGLE_CORRELATION_250', 'CIRCULAR_HULL_HEAD_3POINT_ANGLE_CORRELATION_500', 'CIRCULAR_HULL_HEAD_3POINT_ANGLE_CORRELATION_1000', 'CIRCULAR_HULL_HEAD_3POINT_ANGLE_CORRELATION_2000'])
            hull_tail_angular_corr = pd.DataFrame(CircularStatisticsMixin.sliding_circular_correlation(sample_1=results['CIRCULAR_FRAME_HULL_3POINT_ANGLE'].values.astype(np.float32), sample_2=results['CIRCULAR_FRAME_TAIL_3POINT_ANGLE'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['CIRCULAR_HULL_TAIL_3POINT_ANGLE_CORRELATION_250', 'CIRCULAR_HULL_TAIL_3POINT_ANGLE_CORRELATION_500', 'CIRCULAR_HULL_TAIL_3POINT_ANGLE_CORRELATION_1000', 'CIRCULAR_HULL_TAIL_3POINT_ANGLE_CORRELATION_2000'])
            mean_resultant_vector_length = pd.DataFrame(CircularStatisticsMixin.sliding_mean_resultant_vector_length(data=direction_degrees, fps=int(fps), time_windows=TIME_WINDOWS), columns=['CIRCULAR_MEAN_RESULTANT_LENGTH_250', 'CIRCULAR_MEAN_RESULTANT_LENGTH_500', 'CIRCULAR_MEAN_RESULTANT_LENGTH_1000', 'CIRCULAR_MEAN_RESULTANT_LENGTH_2000'])
            results = pd.concat([results, angular_difference, rao_spacing, circular_range, circular_std, head_hull_angular_corr, hull_tail_angular_corr, mean_resultant_vector_length], axis=1)

            # MOVEMENT FEATURES
            print('Compute movement features...')
            results['MOVEMENT_FRAME_NOSE'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=nose_arr, location_2=shifted_[[f'{NOSE}_x', f'{NOSE}_y']].values.astype(np.float32), px_per_mm=px_per_mm).astype(np.int32)
            results['MOVEMENT_FRAME_CENTER'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=center_arr, location_2=shifted_[[f'{CENTER}_x', f'{CENTER}_y']].values.astype(np.float32), px_per_mm=px_per_mm).astype(np.int32)
            results['MOVEMENT_FRAME_TAILBASE'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=tailbase_arr, location_2=shifted_[[f'{TAIL_BASE}_x', f'{TAIL_BASE}_y']].values.astype(np.float32), px_per_mm=px_per_mm).astype(np.int32)
            results['MOVEMENT_FRAME_TAILTIP'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=tail_tip_arr, location_2=shifted_[[f'{TAIL_TIP}_x', f'{TAIL_TIP}_y']].values.astype(np.float32), px_per_mm=px_per_mm).astype(np.int32)
            results['MOVEMENT_FRAME_TAILCENTER'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=tail_tip_arr, location_2=shifted_[[f'{TAIL_CENTER}_x', f'{TAIL_CENTER}_y']].values.astype(np.float32), px_per_mm=px_per_mm).astype(np.int32)
            results['MOVEMENT_FRAME_LEFT_EAR'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=left_ear_arr, location_2=shifted_[[f'{LEFT_EAR}_x', f'{LEFT_EAR}_y']].values.astype(np.float32), px_per_mm=px_per_mm).astype(np.int32)
            results['MOVEMENT_FRAME_RIGHT_EAR'] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=right_ear_arr, location_2=shifted_[[f'{RIGHT_EAR}_x', f'{RIGHT_EAR}_y']].values.astype(np.float32), px_per_mm=px_per_mm).astype(np.int32)
            results['MOVEMENT_FRAME_SUMMED'] = results['MOVEMENT_FRAME_NOSE'] + results['MOVEMENT_FRAME_CENTER'] + results['MOVEMENT_FRAME_TAILBASE'] + results['MOVEMENT_FRAME_TAILTIP'] + results['MOVEMENT_FRAME_TAILCENTER'] + results['MOVEMENT_FRAME_LEFT_EAR'] + results['MOVEMENT_FRAME_RIGHT_EAR']
            results['MOVEMENT_NOSE_ACCELERATION_MM_S'] = TimeseriesFeatureMixin.acceleration(data=results['MOVEMENT_FRAME_NOSE'].values.astype(np.float32), pixels_per_mm=px_per_mm, fps=fps)
            results['MOVEMENT_CENTER_ACCELERATION_MM_S'] = TimeseriesFeatureMixin.acceleration(data=results['MOVEMENT_FRAME_CENTER'].values.astype(np.float32), pixels_per_mm=px_per_mm, fps=fps)
            results['MOVEMENT_TAILBASE_ACCELERATION_MM_S'] = TimeseriesFeatureMixin.acceleration(data=results['MOVEMENT_FRAME_TAILBASE'].values.astype(np.float32), pixels_per_mm=px_per_mm, fps=fps)
            results['MOVEMENT_TAILTIP_ACCELERATION_MM_S'] = TimeseriesFeatureMixin.acceleration(data=results['MOVEMENT_FRAME_TAILTIP'].values.astype(np.float32), pixels_per_mm=px_per_mm, fps=fps)
            results['MOVEMENT_TAILCENTER_ACCELERATION_MM_S'] = TimeseriesFeatureMixin.acceleration(data=results['MOVEMENT_FRAME_TAILCENTER'].values.astype(np.float32), pixels_per_mm=px_per_mm, fps=fps)
            nose_center_acceleration_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_NOSE_ACCELERATION_MM_S'].values.astype(np.float32), sample_2=results['MOVEMENT_CENTER_ACCELERATION_MM_S'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_NOSE_CENTER_ACCELERATION_SPEARMAN_CORRELATION_250', 'MOVEMENT_NOSE_CENTER_ACCELERATION_SPEARMAN_CORRELATION_500', 'MOVEMENT_NOSE_CENTER_ACCELERATION_SPEARMAN_CORRELATION_1000',  'MOVEMENT_NOSE_CENTER_ACCELERATION_SPEARMAN_CORRELATION_2000'])
            nose_tailbase_acceleration_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_NOSE_ACCELERATION_MM_S'].values.astype(np.float32), sample_2=results['MOVEMENT_TAILBASE_ACCELERATION_MM_S'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_NOSE_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_250', 'MOVEMENT_NOSE_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_500', 'MOVEMENT_NOSE_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_1000',  'MOVEMENT_NOSE_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_2000'])
            center_tailbase_acceleration_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_CENTER_ACCELERATION_MM_S'].values.astype(np.float32), sample_2=results['MOVEMENT_TAILBASE_ACCELERATION_MM_S'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_CENTER_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_250', 'MOVEMENT_CENTER_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_500', 'MOVEMENT_CENTER_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_1000',  'MOVEMENT_CENTER_TAILBASE_ACCELERATION_SPEARMAN_CORRELATION_2000'])
            tailtip_tailbase_acceleration_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_TAILBASE_ACCELERATION_MM_S'].values.astype(np.float32), sample_2=results['MOVEMENT_TAILTIP_ACCELERATION_MM_S'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_TAILBASE_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_250', 'MOVEMENT_TAILBASE_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_500', 'MOVEMENT_TAILBASE_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_1000',  'MOVEMENT_TAILBASE_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_2000'])
            tailcenter_tailend_acceleration_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_TAILCENTER_ACCELERATION_MM_S'].values.astype(np.float32), sample_2=results['MOVEMENT_TAILTIP_ACCELERATION_MM_S'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_TAILCENTER_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_250', 'MOVEMENT_TAILCENTER_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_500', 'MOVEMENT_TAILCENTER_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_1000',  'MOVEMENT_TAILCENTER_TAILEND_ACCELERATION_SPEARMAN_CORRELATION_2000'])
            nose_center_movement_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_FRAME_NOSE'].values.astype(np.float32), sample_2=results['MOVEMENT_FRAME_CENTER'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_NOSE_CENTER_MOVEMENT_SPEARMAN_CORRELATION_250', 'MOVEMENT_NOSE_CENTER_MOVEMENT_SPEARMAN_CORRELATION_500', 'MOVEMENT_NOSE_CENTER_MOVEMENT_SPEARMAN_CORRELATION_1000',  'MOVEMENT_NOSE_CENTER_MOVEMENT_SPEARMAN_CORRELATION_2000'])
            nose_tailbase_movement_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_FRAME_NOSE'].values.astype(np.float32), sample_2=results['MOVEMENT_FRAME_TAILBASE'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_NOSE_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_250', 'MOVEMENT_NOSE_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_500', 'MOVEMENT_NOSE_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_1000',  'MOVEMENT_NOSE_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_2000'])
            center_tailbase_movement_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_FRAME_CENTER'].values.astype(np.float32), sample_2=results['MOVEMENT_FRAME_TAILBASE'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_CENTER_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_250', 'MOVEMENT_CENTER_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_500', 'MOVEMENT_CENTER_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_1000',  'MOVEMENT_CENTER_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_2000'])
            tailbase_tailend_movement_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_FRAME_TAILBASE'].values.astype(np.float32), sample_2=results['MOVEMENT_FRAME_TAILTIP'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_TAILTIP_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_250', 'MOVEMENT_TAILTIP_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_500', 'MOVEMENT_TAILTIP_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_1000',  'MOVEMENT_TAILTIP_TAILBASE_MOVEMENT_SPEARMAN_CORRELATION_2000'])
            tailcenter_tailend_movement_spearman = pd.DataFrame(Statistics.sliding_spearman_rank_correlation(sample_1=results['MOVEMENT_FRAME_TAILCENTER'].values.astype(np.float32), sample_2=results['MOVEMENT_FRAME_TAILTIP'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=fps), columns=['MOVEMENT_TAILTIP_TAILCENTER_MOVEMENT_SPEARMAN_CORRELATION_250', 'MOVEMENT_TAILTIP_TAILCENTER_MOVEMENT_SPEARMAN_CORRELATION_500', 'MOVEMENT_TAILTIP_TAILCENTER_MOVEMENT_SPEARMAN_CORRELATION_1000',  'MOVEMENT_TAILTIP_TAILCENTER_MOVEMENT_SPEARMAN_CORRELATION_2000'])
            results = pd.concat([results, nose_center_acceleration_spearman, nose_tailbase_acceleration_spearman, center_tailbase_acceleration_spearman, tailtip_tailbase_acceleration_spearman, tailcenter_tailend_acceleration_spearman, nose_center_movement_spearman, nose_tailbase_movement_spearman, center_tailbase_movement_spearman, tailbase_tailend_movement_spearman, tailcenter_tailend_movement_spearman], axis=1)

            dominant_f_nose = pd.DataFrame(Statistics.sliding_dominant_frequencies(data=results['MOVEMENT_FRAME_NOSE'].values.astype(np.float32), fps=fps, k=2, time_windows=TIME_WINDOWS), columns=['MOVEMENT_NOSE_MOVEMENT_DOMINANT_FREQUENCY_250', 'MOVEMENT_NOSE_MOVEMENT_DOMINANT_FREQUENCY_500', 'MOVEMENT_NOSE_MOVEMENT_DOMINANT_FREQUENCY_1000', 'MOVEMENT_NOSE_MOVEMENT_DOMINANT_FREQUENCY_2000'])
            dominant_f_center = pd.DataFrame(Statistics.sliding_dominant_frequencies(data=results['MOVEMENT_FRAME_CENTER'].values.astype(np.float32), fps=fps, k=2, time_windows=TIME_WINDOWS), columns=['MOVEMENT_CENTER_MOVEMENT_DOMINANT_FREQUENCY_250', 'MOVEMENT_CENTER_MOVEMENT_DOMINANT_FREQUENCY_500', 'MOVEMENT_CENTER_MOVEMENT_DOMINANT_FREQUENCY_1000', 'MOVEMENT_CENTER_MOVEMENT_DOMINANT_FREQUENCY_2000'])
            results = pd.concat([results, dominant_f_nose, dominant_f_center], axis=1)
            results['MOVEMENT_NOSE_AUTOCORRELATION_500'] = Statistics.sliding_autocorrelation(data=results['MOVEMENT_FRAME_NOSE'].values.astype(np.float32), max_lag=0.5, time_window=1.0, fps=fps)

            for time, bp in product(TIME_WINDOWS, [NOSE, CENTER, 'TAILTIP', 'TAILCENTER', 'SUMMED']):
                results[f'MOVEMENT_MEAN_{time}_{bp.upper()}'] = results[f'MOVEMENT_FRAME_{bp.upper()}'].rolling(int(time * fps), min_periods=1).mean()
                results[f'MOVEMENT_VAR_{time}_{bp.upper()}'] = results[f'MOVEMENT_FRAME_{bp.upper()}'].rolling(int(time * fps), min_periods=1).var()
                results[f'MOVEMENT_SUM_{time}_{bp.upper()}'] = results[f'MOVEMENT_FRAME_{bp.upper()}'].rolling(int(time * fps), min_periods=1).sum()

            # POSE CONFIDENCE FEATURES
            print('Compute probability features...')
            p_df = pd.DataFrame(FeatureExtractionMixin.count_values_in_range(data=p_arr, ranges=np.array([[0.0, 0.25], [0.25, 0.50], [0.50, 0.75], [0.75, 1.0]])), columns=['PROBABILITIES_LOW_COUNT', 'PROBABILITIES_MEDIUM_LOW_COUNT', 'PROBABILITIES_MEDIUM_HIGHT', 'PROBABILITIES_HIGH_COUNT']).astype(np.int32)
            sliding_z_p_low = pd.DataFrame(Statistics.sliding_z_scores(data=p_df['PROBABILITIES_LOW_COUNT'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=int(fps)), columns=[f'PROBABILITIES_LOW_COUNT_SLIDING_Z_SCORE_250', f'PROBABILITIES_LOW_COUNT_SLIDING_Z_SCORE_500', f'PROBABILITIES_LOW_COUNT_SLIDING_Z_SCORE_1000', f'PROBABILITIES_LOW_COUNT_SLIDING_Z_SCORE_2000'])
            sliding_z_p_high = pd.DataFrame(Statistics.sliding_z_scores(data=p_df['PROBABILITIES_HIGH_COUNT'].values.astype(np.float32), time_windows=TIME_WINDOWS, fps=int(fps)), columns=[f'PROBABILITIES_HIGH_COUNT_SLIDING_Z_SCORE_250', f'PROBABILITIES_HIGH_COUNT_SLIDING_Z_SCORE_500', f'PROBABILITIES_HIGH_COUNT_SLIDING_Z_SCORE_1000', f'PROBABILITIES_HIGH_COUNT_SLIDING_Z_SCORE_2000'])
            results = pd.concat([df, results, p_df, sliding_z_p_low, sliding_z_p_high], axis=1).fillna(-1)
            self.save(df=results, save_path=save_path)
            video_timer.stop_timer()
            print(f'Video {video_name} complete (elapsed time: {video_timer.elapsed_time_str}s)...')

        self.timer.stop_timer()
        stdout_success(msg=f'Features extracted for {len(self.outlier_corrected_paths)} files(s)', elapsed_time=self.timer.elapsed_time_str)

    def save(self, df: pd.DataFrame, save_path: os.PathLike):
        write_df(df=df.astype(np.float32), file_type=self.file_type, save_path=save_path)

# if __name__ == "__main__":
#     parser = argparse.ArgumentParser(description='SimBA Custom Feature Extractor')
#     parser.add_argument('--config_path', type=str, help='SimBA project config path')
#     args = parser.parse_args()
#     feature_extractor = MitraFeatureExtractor(config_path=args.config_path)
#     feature_extractor.run()
#



# feature_extractor = MitraFeatureExtractor(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini")
# feature_extractor.run()











__author__ = "Simon Nilsson"

import os
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.directing_animals_visualizer import \
    DirectingOtherAnimalsVisualizer
from simba.plotting.directing_animals_visualizer_mp import \
    DirectingOtherAnimalsVisualizerMultiprocess
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        SimbaButton)
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import AnimalNumberError
from simba.utils.lookups import get_color_dict
from simba.utils.read_write import find_all_videos_in_directory

DIRECTION_THICKNESS = "direction_thickness"
DIRECTIONALITY_COLOR = "directionality_color"
CIRCLE_SIZE = "circle_size"
HIGHLIGHT_ENDPOINTS = "highlight_endpoints"
SHOW_POSE = "show_pose"
ANIMAL_NAMES = "animal_names"


class DirectingOtherAnimalsVisualizerPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if self.animal_cnt == 1:
            raise AnimalNumberError(msg="Cannot visualize directionality between animals in a 1 animal project.", source=self.__class__.__name__,)
        PopUpMixin.__init__(self, title="CREATE ANIMAL DIRECTION VIDEOS", icon='eye')
        self.color_dict = get_color_dict()
        self.color_lst = list(self.color_dict.keys())
        self.color_lst.insert(0, "random")
        self.size_lst = list(range(1, 11))
        self.files_found_dict = find_all_videos_in_directory(directory=self.video_dir, as_dict=True)
        self.show_pose_var = BooleanVar(value=True)
        self.show_animal_names_var = BooleanVar(value=True)
        self.highlight_direction_endpoints_var = BooleanVar(value=True)
        self.multiprocess_var = BooleanVar(value=False)

        self.style_settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="STYLE SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.DIRECTING_ANIMALS_PLOTS.value,)
        self.show_pose_cb = Checkbutton(self.style_settings_frm, text="Show pose-estimated body-parts", font=Formats.FONT_REGULAR.value, variable=self.show_pose_var,)
        self.highlight_direction_endpoints_cb = Checkbutton( self.style_settings_frm, text="Highlight direction end-points", font=Formats.FONT_REGULAR.value, variable=self.highlight_direction_endpoints_var,)
        self.show_animal_names_cb = Checkbutton( self.style_settings_frm, text="Show animal names", font=Formats.FONT_REGULAR.value, variable=self.show_animal_names_var,)

        self.direction_clr_dropdown = DropDownMenu(self.style_settings_frm, "Direction color:", self.color_lst, "16")
        self.pose_size_dropdown = DropDownMenu(self.style_settings_frm, "Pose circle size:", self.size_lst, "16")
        self.line_thickness = DropDownMenu(self.style_settings_frm, "Line thickness:", self.size_lst, "16")
        self.line_thickness.setChoices(choice=4)
        self.pose_size_dropdown.setChoices(choice=3)
        self.direction_clr_dropdown.setChoices(choice="random")
        multiprocess_cb = Checkbutton(
            self.style_settings_frm,
            font=Formats.FONT_REGULAR.value,
            text="Multi-process (faster)",
            variable=self.multiprocess_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocess_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown = DropDownMenu(self.style_settings_frm, "CPU cores:", list(range(2, self.cpu_cnt)), "12")
        self.multiprocess_dropdown.setChoices(2)
        self.multiprocess_dropdown.disable()

        self.run_frm = LabelFrame(self.main_frm,text="RUN",font=Formats.FONT_HEADER.value,pady=5,padx=5,fg="black")
        self.run_single_video_frm = LabelFrame( self.run_frm, text="SINGLE VIDEO", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black",)


        self.run_single_video_btn = SimbaButton(parent=self.run_single_video_frm, txt="Create single video", font=Formats.FONT_REGULAR.value, cmd=self.__create_directionality_plots, cmd_kwargs={'multiple_videos': False})
        self.single_video_dropdown = DropDownMenu(self.run_single_video_frm,"Video:",list(self.files_found_dict.keys()),"12",)
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])
        self.run_multiple_videos = LabelFrame(self.run_frm,text="MULTIPLE VIDEO",font=Formats.FONT_HEADER.value,pady=5,padx=5,fg="black",)

        self.run_multiple_video_btn = SimbaButton(parent=self.run_multiple_videos, txt=f"Create multiple videos ({len(list(self.files_found_dict.keys()))} video(s) found)", font=Formats.FONT_REGULAR.value, cmd=self.__create_directionality_plots, cmd_kwargs={'multiple_videos': True})
        self.style_settings_frm.grid(row=0, column=0, sticky=NW)
        self.show_pose_cb.grid(row=0, column=0, sticky=NW)
        self.highlight_direction_endpoints_cb.grid(row=1, column=0, sticky=NW)
        self.show_animal_names_cb.grid(row=2, column=0, sticky=NW)
        self.direction_clr_dropdown.grid(row=3, column=0, sticky=NW)
        self.pose_size_dropdown.grid(row=4, column=0, sticky=NW)
        self.line_thickness.grid(row=5, column=0, sticky=NW)
        multiprocess_cb.grid(row=6, column=0, sticky=NW)
        self.multiprocess_dropdown.grid(row=6, column=1, sticky=NW)

        self.run_frm.grid(row=1, column=0, sticky=NW)
        self.run_single_video_frm.grid(row=0, column=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)
        self.run_multiple_videos.grid(row=1, column=0, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, column=0, sticky=NW)

        self.main_frm.mainloop()

    def __create_directionality_plots(self, multiple_videos: bool):
        if multiple_videos:
            video_paths = list(self.files_found_dict.values())
        else:
            video_paths = [
                self.files_found_dict[self.single_video_dropdown.getChoices()]
            ]

        direction_clr = self.direction_clr_dropdown.getChoices()
        if direction_clr != "random":
            direction_clr = self.color_dict[direction_clr]

        style_attr = {
            SHOW_POSE: self.show_pose_var.get(),
            CIRCLE_SIZE: int(self.pose_size_dropdown.getChoices()),
            DIRECTIONALITY_COLOR: direction_clr,
            DIRECTION_THICKNESS: int(self.line_thickness.getChoices()),
            HIGHLIGHT_ENDPOINTS: self.highlight_direction_endpoints_var.get(),
            ANIMAL_NAMES: self.show_animal_names_var.get(),
        }

        for video_path in video_paths:
            if not self.multiprocess_var.get():
                visualizer = DirectingOtherAnimalsVisualizer(
                    config_path=self.config_path,
                    video_path=video_path,
                    style_attr=style_attr,
                )
            else:
                visualizer = DirectingOtherAnimalsVisualizerMultiprocess(
                    config_path=self.config_path,
                    video_path=video_path,
                    style_attr=style_attr,
                    core_cnt=int(self.multiprocess_dropdown.getChoices()),
                )

            threading.Thread(target=visualizer.run()).start()


#_ = DirectingOtherAnimalsVisualizerPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# _ = DirectingOtherAnimalsVisualizerPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')



import os
from typing import Tuple, Union

import numpy as np
import pandas as pd

from simba.utils.checks import (check_float, check_if_dir_exists, check_int,
                                check_valid_array)
from simba.utils.printing import SimbaTimer, stdout_success


class GibbSampler():

    """
     Gibbs sampling for finding "motifs" in categorical sequences.

     :param np.ndarray data: 2-dimensional array where observations are organised by row and each sequential sample in the observation is organized by column.
     :param Union[str, os.PathLike] save_path: The path location where to save the CSV results.
     :param int sequence_length: The length of the motif sequence searched for.
     :param int iterations: Number of iterations per epoch. Default: 1500.
     :param int epochs: Number of epochs of iterations. Default: 4.
     :param float stop_val: Terminate once the error value reaches below this threshold. Default 0.001.
     :param int plateau_val: Terminate epoch when error rate has remained unchanged for ``plateau_val`` count of iterations. Default 50.
     :param float pseudo_number: Small error value for fuzzy search and avoid division by zero errors. Default: 10e-6.

     :example:
     >>> data = pd.read_csv(r"/Users/simon/Desktop/envs/simba/simba/tests/data/sample_data/gibbs_sample_cardinal.csv", index_col=0).values
     >>> sampler = GibbSampler(data=data, save_path=r'/Users/simon/Desktop/gibbs.csv', epochs=5, iterations=600)
     >>> sampler.run()

     :references:
        .. [1] Lawrence et al, Detecting Subtle Sequence Signals: a Gibbs Sampling Strategy for Multiple Alignment, Science, vol. 262, pp. 208-214, 1993.
        .. [2] Great YouTube tutorial / explanation by Xiaole Shirley Liu - `https://www.youtube.com/watch?v=NRjhfyXWHuQ <https://www.youtube.com/watch?v=NRjhfyXWHuQ>`_.
        .. [3] Weinreb et al. Keypoint-MoSeq: parsing behavior by linking point tracking to pose dynamics, Nature Methods, 21, 1329–1339 (2024).
     """

    def __init__(self,
                 data: np.ndarray,
                 save_path: Union[str, os.PathLike],
                 sequence_length: int = 4,
                 iterations: int = 1500,
                 epochs: int = 2,
                 stop_val: float = 0.001,
                 pseudo_number: float = 10e-6,
                 plateau_val: int = 50):

        check_valid_array(data=data, source=f'{self.__class__.__name__} data', accepted_ndims=(2,))
        check_float(name=f'{self.__class__.__name__} stop_val', value=stop_val, min_value=0.0)
        check_float(name=f'{self.__class__.__name__} pseudo_number', value=pseudo_number, min_value=10e-16)
        check_int(name=f'{self.__class__.__name__} epochs', value=epochs, min_value=1)
        check_int(name=f'{self.__class__.__name__} iterations', value=iterations, min_value=1)
        check_int(name=f'{self.__class__.__name__} sequence_length', value=sequence_length, min_value=2)
        check_if_dir_exists(in_dir=os.path.dirname(save_path), source=f'{self.__class__.__name__} save_path')
        self.unique_vals = np.sort(np.unique(data.flatten()))
        self.target_probability = 1 * sequence_length + (pseudo_number * (sequence_length + 1))
        self.holdout_fields = [f"H_{i}" for i in range(sequence_length)]
        self.non_holdout_cols = [f"C_{i}" for i in range(sequence_length)]
        self.out_cols = [f"Behavior_{i+1}" for i in range(sequence_length)]
        self.summary_df = pd.DataFrame(columns=self.out_cols)
        self.data, self.pseudo_num, self.sequence_len, self.plateau_val = (data, pseudo_number, sequence_length, plateau_val)
        self.iterations, self.stop_val, self.epochs = [iterations] * epochs, stop_val, epochs
        self.save_path = save_path

    def __make_probability_table(self, data: np.ndarray) -> pd.DataFrame:
        prob_df = pd.DataFrame(columns=self.unique_vals)
        for field_idx in range(data.shape[1]):
            idx_data = data[:, field_idx].flatten()
            for unique_val in self.unique_vals:
                val_idx = np.argwhere(idx_data == unique_val).flatten()
                pct = (val_idx.shape[0] + self.pseudo_num) / (idx_data.shape[0] + self.pseudo_num)
                prob_df.loc[field_idx, unique_val] = pct
        return prob_df

    def __iterate_through_possible_seq_in_hold_out(self,
                                                   holdout_obs: np.ndarray,
                                                   bg_dict: dict,
                                                   probability_df: pd.DataFrame) -> pd.DataFrame:

        out_df = pd.DataFrame(columns=self.holdout_fields + ["weight"])
        for r in range(self.sequence_len, holdout_obs.shape[0]+1):
            sequence = holdout_obs[r - self.sequence_len:r]
            prob_tot = 1
            for i in range(self.sequence_len):
                val = sequence[i]
                prob_tot *= probability_df.loc[i][val] / bg_dict[val]
            out_df.loc[len(out_df)] = np.append(sequence, prob_tot)
        return out_df

    def __sum_results(self, full_sequence_set, summary_df) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_sequence_set = pd.DataFrame(full_sequence_set, columns=self.out_cols)
        summary_df = pd.concat([summary_df, full_sequence_set], axis=0).reset_index(drop=True)
        output = (summary_df.groupby(summary_df.columns.tolist()).size().reset_index().rename(columns={0: "records"}).sort_values(by=["records"], ascending=False))
        output["percent"] = output["records"] / output["records"].sum()
        return summary_df, output.reset_index(drop=True)

    def run(self):
        timer = SimbaTimer(start=True)
        unique_elements, counts = np.unique(self.data, return_counts=True)
        counts_dict = dict(zip(unique_elements, counts))
        bg_dict = {k: (v / np.sum(counts)) for k, v in counts_dict.items()}
        prior_error, stable_error_cnt, full_sequence, output = np.inf, 0, None, None

        for epoch_cnt, iterations in enumerate(self.iterations):
            holdout_idx = np.random.randint(0, self.data.shape[0]-1)
            for iteration in range(0, iterations):
                hold_out_obs = self.data[holdout_idx]
                if iteration == 0:
                    start_pos = np.random.randint(low=0, high=self.data.shape[1] - self.sequence_len, size=(self.data.shape[0]))
                    end_pos = np.array([x + self.sequence_len for x in start_pos])
                    slices = np.array([[x, y] for x, y in zip(start_pos, end_pos)])
                    epoch_sample = np.delete(self.data, [holdout_idx], axis=0)
                    sequences = np.full(shape=(epoch_sample.shape[0], self.sequence_len), fill_value='-1.0')
                    for idx in range(epoch_sample.shape[0]):
                        i = slices[idx]
                        sequences[idx] = epoch_sample[idx][i[0]: i[1]]
                else:
                    sequences = np.delete(full_sequence, [holdout_idx], axis=0)
                probability_df = self.__make_probability_table(sequences)
                df_holdout_seqs = self.__iterate_through_possible_seq_in_hold_out(hold_out_obs, bg_dict, probability_df)
                new_sequence = df_holdout_seqs.sample(1, weights="weight")[self.holdout_fields].values[0]
                full_sequence = np.insert(sequences, 0, new_sequence, axis=0)
                error = round(self.target_probability - probability_df.max(axis=1).sum(), 4)
                print(f"Current error: {error}, iteration: {iteration+1}/{iterations}, epoch: {epoch_cnt+1}/{self.epochs}")
                if error <= self.stop_val:
                    print(f"Convergence reached. Error: {error}, Stop value: {self.stop_val}")
                    break
                if error == prior_error:
                    stable_error_cnt += 1
                    if stable_error_cnt >= self.plateau_val - 1:
                        print(f"Plateau reached. Stable error count: {stable_error_cnt}. Plateau value: {self.plateau_val}")
                        break
                else:
                    stable_error_cnt = 0
                prior_error = error
                if holdout_idx < (len(self.data) - 1):
                    holdout_idx += 1
                else:
                    holdout_idx = 0
                self.summary_df, output = self.__sum_results(full_sequence, self.summary_df)

        output.to_csv(self.save_path)
        timer.stop_timer()
        stdout_success(msg=f"Gibbs sampling results saved @{self.save_path}", elapsed_time=timer.elapsed_time_str)



# data = pd.read_csv(r"C:\projects\simba\simba\tests\data\sample_data\gibbs_sample_cardinal.csv", index_col=0).values
# sampler = GibbSampler(data=data, save_path=r'C:\Users\sroni\OneDrive\Desktop\gibbs.csv', iterations=5)
# sampler.run()

# data = pd.read_csv(r"/Users/simon/Desktop/envs/simba/simba/tests/data/sample_data/gibbs_sample_cardinal.csv", index_col=0).values
# sampler = GibbSampler(data=data, save_path=r'/Users/simon/Desktop/gibbs.csv', epochs=5, iterations=600)
# sampler.run()


__author__ = "Simon Nilsson"

import os
from tkinter import *
from typing import Optional, Union

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.px_to_mm_ui import GetPixelsPerMillimeterInterface
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimbaButton, SimBALabel)
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_int)
from simba.utils.enums import (ConfigKey, Dtypes, Formats, Keys, Links,
                               Options, TagNames)
from simba.utils.errors import InvalidInputError, PermissionError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_video_meta_data, read_config_entry,
                                    read_video_info_csv)

TABLE_HEADERS = ["INDEX", "VIDEO", "FPS", "RESOLUTION WIDTH", "RESOLUTION HEIGHT", "FIND DISTANCE", "DISTANCE IN MM", "PIXELS PER MM"]


class VideoInfoTable(ConfigReader, PopUpMixin):
    """
    Create GUI that allows users to modify resolutions, fps, and pixels-per-mm
    interactively of videos within the SimBA project. Data is stored within the project_folder/logs/video_info.csv
    file in the SimBA project.

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] video_dir: Optional path to directory with video files. If None, then read from the SimBA project as dictated by project config. Default None.
    :param Optional[Union[str, os.PathLike]] video_info_path: Optional path to vide_info.csv file. If None, then read from the SimBA project as dictated by project config. Default None.

    ..seealso::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-3-set-video-parameters>`__.

    :example:
    >>> ui = VideoInfoTable(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
    >>> ui.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 video_dir: Optional[Union[str, os.PathLike]] = None,
                 video_info_path: Optional[Union[str, os.PathLike]] = None):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        if video_dir is not None:
            check_if_dir_exists(in_dir=video_dir, source=self.__class__.__name__)
            self.video_dir = video_dir
        if video_info_path is not None:
            check_file_exist_and_readable(file_path=video_info_path)
            self.video_info_path = video_info_path
        if os.path.isfile(self.video_info_path):
            self.video_info_df = read_video_info_csv(self.video_info_path).reset_index(drop=True)
            self.prior_videos = list(self.video_info_df['Video'])
        else:
            self.video_info_df, self.prior_videos = None, []
        self.distance_mm = read_config_entry(self.config, ConfigKey.FRAME_SETTINGS.value, ConfigKey.DISTANCE_MM.value, Dtypes.FLOAT.value,0.00)
        self.video_paths = find_files_of_filetypes_in_directory(directory=self.video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True, as_dict=True)
        self.max_char_vid_name = len(max(self.video_paths.keys(), key=len))
        PopUpMixin.__init__(self, title="VIDEO INFO", size=(1550, 800), icon='video')
        self.TABLE_COL_WIDTHS = ["5", self.max_char_vid_name + 5, "18", "18", "18", "18", "18", "18"]


    def _get_video_table_rows(self):
        self.videos = {}
        for cnt, (video_name, video_path) in enumerate(self.video_paths.items()):
            self.videos[video_name] = {}
            if video_name in self.prior_videos:
                try:
                    prior_data = self.read_video_info(video_name=video_name, raise_error=False)[0].reset_index(drop=True).iloc[0].to_dict()
                    fps, width, height, distance, pixels_per_mm = prior_data['fps'], prior_data['Resolution_width'], prior_data['Resolution_height'], prior_data['Distance_in_mm'], prior_data['pixels/mm']
                except:
                    fps, width, height = self.video_meta_data[video_name]['fps'], self.video_meta_data[video_name]['width'], self.video_meta_data[video_name]['height']
                    distance, pixels_per_mm = self.distance_mm, 'None'
            else:
                fps, width, height = self.video_meta_data[video_name]['fps'], self.video_meta_data[video_name]['width'], self.video_meta_data[video_name]['height']
                distance, pixels_per_mm = self.distance_mm, 'None'
            self.videos[video_name]["video_idx_lbl"] = Label(self.video_frm, text=str(cnt), font=Formats.FONT_REGULAR.value,  width=6)
            self.videos[video_name]["video_name_lbl"] = Label(self.video_frm, text=video_name, font=Formats.FONT_REGULAR.value,  width=self.max_char_vid_name)
            self.videos[video_name]["video_name_w_ext"] = os.path.basename(video_path)
            self.videos[video_name]["video_fps_eb"] = Entry_Box(parent=self.video_frm, fileDescription='', labelwidth=0, value=fps, entry_box_width=20, justify='center')
            self.videos[video_name]["video_width_eb"] = Entry_Box(parent=self.video_frm, fileDescription='', labelwidth=0, value=width, entry_box_width=20, justify='center', validation='numeric')
            self.videos[video_name]["video_height_eb"] = Entry_Box(parent=self.video_frm, fileDescription='', labelwidth=0, value=height, entry_box_width=20, justify='center', validation='numeric')
            self.videos[video_name]["video_known_distance_eb"] = Entry_Box(parent=self.video_frm, fileDescription='', labelwidth=0, value=distance, entry_box_width=20, justify='center')
            #self.videos[video_name]["find_dist_btn"] = Button(self.video_frm, text="CALCULATE DISTANCE", fg="black", command=lambda k=video_name: self._initate_find_distance(k))
            self.videos[video_name]["find_dist_btn"] = SimbaButton(parent=self.video_frm, txt="CALCULATE DISTANCE", txt_clr='black', img='calipher', font=Formats.FONT_HEADER.value, cmd=lambda k=video_name: self._initate_find_distance(k), hover_font=Formats.FONT_HEADER.value)
            self.videos[video_name]["video_px_per_mm"] = Entry_Box(parent=self.video_frm, fileDescription='', labelwidth=0, value=pixels_per_mm, entry_box_width=20, justify='center')
            self.videos[video_name]["video_idx_lbl"].grid(row=cnt+2, column=0, sticky=NW)
            self.videos[video_name]["video_name_lbl"].grid(row=cnt+2, column=1, sticky=NW)
            self.videos[video_name]["video_fps_eb"].grid(row=cnt+2, column=2, sticky=NW)
            self.videos[video_name]["video_width_eb"].grid(row=cnt+2, column=3, sticky=NW)
            self.videos[video_name]["video_height_eb"].grid(row=cnt+2, column=4, sticky=NW)
            self.videos[video_name]["find_dist_btn"].grid(row=cnt + 2, column=5, sticky=NW)
            self.videos[video_name]["video_known_distance_eb"].grid(row=cnt+2, column=6, sticky=NW)
            self.videos[video_name]["video_px_per_mm"].grid(row=cnt+2, column=7, sticky=NW)

    def _get_quick_set(self):
        self.quick_set_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="BATCH QUICK SET", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.VIDEO_PARAMETERS.value)
        self.quick_set_frm_known_distance_eb = Entry_Box(parent=self.quick_set_frm, fileDescription='KNOWN DISTANCE:', labelwidth=30, value='', entry_box_width=10)
        self.quick_set_frm_known_distance_btn = SimbaButton(parent=self.quick_set_frm, txt="APPLY", txt_clr='black', img='tick', font=Formats.FONT_HEADER.value, cmd=self._set_known_distance, cmd_kwargs={'distance': lambda: self.quick_set_frm_known_distance_eb.entry_get})
        self.quick_set_px_mm_eb = Entry_Box(parent=self.quick_set_frm, fileDescription='PIXEL PER MILLIMETER:', labelwidth=30, value='', entry_box_width=10)
        self.quick_set_px_mm_btn = SimbaButton(parent=self.quick_set_frm, txt="APPLY", txt_clr='black', img='tick', font=Formats.FONT_HEADER.value, cmd=self._set_px_per_mm, cmd_kwargs={'px_per_mm': lambda: self.quick_set_px_mm_eb.entry_get})
        self.quick_set_frm.grid(row=1, column=0, sticky=NW)
        self.quick_set_frm_known_distance_eb.grid(row=0, column=0, sticky=NW)
        self.quick_set_frm_known_distance_btn.grid(row=0, column=1, sticky=NW)
        self.quick_set_px_mm_eb.grid(row=1, column=0, sticky=NW)
        self.quick_set_px_mm_btn.grid(row=1, column=1, sticky=NW)


    def _set_known_distance(self, distance: str):
        check_float(name=f'{self.__class__.__name__} KNOWN DISTANCE', value=distance, min_value=10e-6, raise_error=True)
        for cnt, (video_name, video_path) in enumerate(self.video_paths.items()):
            self.videos[video_name]["video_known_distance_eb"].entry_set(val=distance)

    def _set_px_per_mm(self, px_per_mm: str):
        check_float(name=f'{self.__class__.__name__} PIXEL PER MILLIMETER', value=px_per_mm, min_value=10e-6, raise_error=True)
        for cnt, (video_name, video_path) in enumerate(self.video_paths.items()):
            self.videos[video_name]["video_px_per_mm"].entry_set(val=px_per_mm)

    def _get_video_meta_data(self):
        self.video_meta_data = {}
        for cnt, (video_name, video_path) in enumerate(self.video_paths.items()):
            self.video_meta_data[video_name] = get_video_meta_data(video_path=video_path)

    def _duplicate_idx_1_distance(self):
        first_key = list(self.video_paths.keys())[0]
        first_val = self.videos[first_key]["video_known_distance_eb"].entry_get
        if not check_float(name=self.__class__.__name__, value=first_val, min_value=10e-6, raise_error=False)[0]:
            raise InvalidInputError(msg=f'Distance for {first_key} cannot be duplicated. Not a valid distance value: {first_val}', source=self.__class__.__name__)
        for cnt, (video_name, video_path) in enumerate(self.video_paths.items()):
            self.videos[video_name]["video_known_distance_eb"].entry_set(val=first_val)

    def _duplicate_idx_1_px_mm(self):
        first_key = list(self.video_paths.keys())[0]
        first_val = self.videos[first_key]["video_px_per_mm"].entry_get
        if not check_float(name=self.__class__.__name__, value=first_val, min_value=10e-6, raise_error=False)[0]:
            raise InvalidInputError(msg=f'Pixel per millimeter for {first_key} cannot be duplicated. Not a valid distance value: {first_val}', source=self.__class__.__name__)
        for cnt, (video_name, video_path) in enumerate(self.video_paths.items()):
            self.videos[video_name]["video_px_per_mm"].entry_set(val=first_val)

    def _initate_find_distance(self, video_name: str):
        video_path = self.video_paths[video_name]
        known_distance = self.videos[video_name]['video_known_distance_eb'].entry_get
        check_float(name=f'{video_name} known distance', value=known_distance, min_value=10e-16, raise_error=True)
        px_per_mm_interface = GetPixelsPerMillimeterInterface(video_path=video_path, known_metric_mm=float(known_distance))
        px_per_mm_interface.run()
        self.videos[video_name]["video_px_per_mm"].entry_set(str(round(px_per_mm_interface.ppm, 5)))


    def _get_instructions_frm(self):
        self.instructions_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="INSTRUCTIONS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.VIDEO_PARAMETERS.value)
        self.instructions_label_1 = SimBALabel(parent=self.instructions_frm, txt='1. Enter the known distance (mm) in "DISTANCE IN MM." Use "autopopulate" or "batch quick set" for similar videos.', font=Formats.FONT_REGULAR.value)
        self.instructions_label_2 = SimBALabel(parent=self.instructions_frm, txt='2. Click on "CALCULTE DISTANCE" button(s) to calculate pixels/mm for each video.', font=Formats.FONT_REGULAR.value)
        self.instructions_label_3 = SimBALabel(parent=self.instructions_frm, txt="3. Click <SAVE PIXEL PER MILLIMETER DATA> when all the data are filled in.", font=Formats.FONT_REGULAR.value)
        self.instructions_frm.grid(row=0, column=0, sticky=W)
        self.instructions_label_1.grid(row=0, column=0, sticky=W)
        self.instructions_label_2.grid(row=1, column=0, sticky=W)
        self.instructions_label_3.grid(row=2, column=0, sticky=W)

    def _get_save_frm(self, index: int):
        self.save_frm = LabelFrame(self.main_frm, text="SAVE", font=Formats.FONT_HEADER.value)
        self.save_data_btn = SimbaButton(parent=self.save_frm, txt="SAVE PIXEL PER MILLIMETER DATA", txt_clr='black', img='save_large', font=Formats.FONT_HEADER.value, cmd=self._save)
        self.save_frm.grid(row=index, column=0, sticky=NW)
        self.save_data_btn.grid(row=0, column=0, sticky=NW)

    def _get_video_table(self):
        self.video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="PROJECT VIDEOS", icon_name='video_large', icon_link=Links.VIDEO_PARAMETERS.value, font=Formats.FONT_HEADER.value)
        for cnt, (col_name, col_width) in enumerate(zip(TABLE_HEADERS, self.TABLE_COL_WIDTHS)):
            col_header_label = Label( self.video_frm, text=col_name, width=col_width, font=Formats.FONT_HEADER.value)
            col_header_label.grid(row=0, column=cnt, sticky=W)
        self.video_frm.grid(row=3, column=0, sticky=NW)
        self._get_video_table_rows()

    def _get_duplicate_btns(self):
        self.duplicate_distance_btn = SimbaButton(parent=self.video_frm, txt="DUPLICATE INDEX 1", txt_clr='red', img='danger', font=Formats.FONT_REGULAR.value, cmd=self._duplicate_idx_1_distance)
        self.duplicate_distance_btn.grid(row=1, column=6, sticky=W, padx=5)
        self.duplicate_btn = SimbaButton(parent=self.video_frm, txt="DUPLICATE INDEX 1", txt_clr='red', img='danger', font=Formats.FONT_REGULAR.value, cmd=self._duplicate_idx_1_px_mm)
        self.duplicate_btn.grid(row=1, column=7, sticky=W, padx=5)


    def _save(self):
        self.video_info_df = pd.DataFrame(columns=Formats.EXPECTED_VIDEO_INFO_COLS.value)
        for video_name, video_path in self.videos.items():
            fps = self.videos[video_name]["video_fps_eb"].entry_get
            width = self.videos[video_name]["video_width_eb"].entry_get
            height = self.videos[video_name]["video_height_eb"].entry_get
            known_distance = self.videos[video_name]["video_known_distance_eb"].entry_get
            px_per_mm = self.videos[video_name]["video_px_per_mm"].entry_get
            check_float(name=f'{video_name} fps', value=fps, min_value=10e-16)
            check_int(name=f'{video_name} width', value=width, min_value=1, raise_error=True)
            check_int(name=f'{video_name} height', value=height, min_value=1, raise_error=True)
            check_float(name=f'{video_name} known_distance', value=known_distance, min_value=10e-16, raise_error=True)
            check_float(name=f'{video_name} px_per_mm', value=px_per_mm, min_value=10e-16, raise_error=True)
            self.video_info_df.loc[len(self.video_info_df)] = [video_name, fps, width, height, known_distance, px_per_mm]
        self.video_info_df.drop_duplicates(subset=["Video"], inplace=True)
        self.video_info_df = self.video_info_df.set_index("Video")
        try:
            self.video_info_df.to_csv(self.video_info_path)
        except PermissionError:
            raise PermissionError(msg=f"SimBA tried to write to {self.video_info_path}, but was not allowed. If this file is open in another program, try closing it.", source=self.__class__.__name__)
        stdout_success(msg=f"Video info saved at {self.video_info_path}", source=self.__class__.__name__)

    def run(self):
        self._get_video_meta_data()
        self._get_instructions_frm()
        self._get_save_frm(index=2)
        self._get_quick_set()
        self._get_video_table()
        self._get_duplicate_btns()
        self._get_save_frm(index=4)


# ui = VideoInfoTable(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# ui.run()



# test.create_window()
#test.main_frm.mainloop()

# test = VideoInfoTable(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini')
# test.create_window()
# test.main_frm.mainloop()

#
# test = VideoInfoTable(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals_2/project_folder/project_config.ini')
# test.create_window()
# test.main_frm.mainloop()

# test = VideoInfoTable(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.create_window()
# test.main_frm.mainloop()



__author__ = "Simon Nilsson"

import os
import threading
import time
from tkinter import *
from tkinter.ttk import Combobox
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.roi_plotter import ROIPlotter
from simba.plotting.roi_plotter_mp import ROIPlotMultiprocess
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, FileSelect, SimbaButton,
                                        SimbaCheckbox, SimBADropDown,
                                        SimBALabel)
from simba.utils.checks import check_file_exist_and_readable, check_float
from simba.utils.enums import ROI_SETTINGS, Formats, Keys, Links, Options
from simba.utils.errors import NoDataError, ROICoordinatesNotFoundError
from simba.utils.lookups import get_color_dict
from simba.utils.read_write import (find_all_videos_in_directory,
                                    find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    str_2_bool)

BP_SIZE_OPTIONS = list(range(1, 101, 1))
BP_SIZE_OPTIONS.insert(0, 'AUTO')

class VisualizeROITrackingPopUp(PopUpMixin, ConfigReader):

    """
    :example:
    >>> VisualizeROITrackingPopUp(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")
    """
    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path, source=self.__class__.__name__)
        if len(self.outlier_corrected_paths) == 0:
            raise NoDataError(msg=f'No data found in {self.outlier_corrected_dir} directory. Make sure the video pose-estimation data is represented in the {self.outlier_corrected_dir} directory')
        self.read_roi_data()
        self.video_file_paths = find_all_videos_in_directory(directory=self.video_dir, as_dict=True, raise_error=True)
        self.data_file_paths = find_files_of_filetypes_in_directory(directory=self.outlier_corrected_dir, extensions=[f'.{self.file_type}'], as_dict=True)
        self.videos_with_rois_and_data = [x for x in self.video_file_paths.keys() if x in list(self.data_file_paths.keys()) and x in self.video_names_w_rois]
        if len(self.videos_with_rois_and_data) == 0:
            raise NoDataError(msg=f'No ROIs found for the videos represented in the {self.video_dir} and data represented in the {self.outlier_corrected_dir} directory. Draw ROIs for these videos before visualizing the data.')
        self.clr_dict = get_color_dict()
        self.longest_animal_name_len = len(max(self.multi_animal_id_list, key=len)) + 5
        gpu_state = NORMAL if self.gpu_available else DISABLED

        PopUpMixin.__init__(self, title="VISUALIZE ROI TRACKING", size=(800, 500), icon='shapes_small')
        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.ROI_DATA_PLOT.value)
        self.threshold_entry_box = Entry_Box(self.settings_frm, "BODY-PART PROBABILITY THRESHOLD:", "35", value='0.0', justify='center', entry_box_width=35)
        self.show_pose_estimation_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=['TRUE', 'FALSE'], dropdown_width=self.longest_animal_name_len, label='SHOW POSE-ESTIMATED LOCATIONS:', label_width=35, value='TRUE', command=self._disable_clr)
        self.show_animal_name_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=['TRUE', 'FALSE'], dropdown_width=self.longest_animal_name_len, label='SHOW ANIMAL NAMES:', label_width=35, value='FALSE')
        self.outside_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=['TRUE', 'FALSE'], dropdown_width=self.longest_animal_name_len, label='OUTSIDE ROI ZONES DATA:', label_width=35, value='FALSE', tooltip_txt=f'TREAT ALL NON-ROI REGIONS AS AN ROI REGION NAMED \n "{ROI_SETTINGS.OUTSIDE_ROI.value}"')
        self.core_cnt_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=list(range(1, self.cpu_cnt)), dropdown_width=self.longest_animal_name_len, label='NUMBER OF CPU CORES:', label_width=35, value=find_core_cnt()[1])
        self.gpu_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=['TRUE', 'FALSE'], dropdown_width=self.longest_animal_name_len, label='USE GPU:', label_width=35, value='FALSE', state=gpu_state)

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.threshold_entry_box.grid(row=0, column=0, sticky=NW)
        self.show_pose_estimation_dropdown.grid(row=2, column=0, sticky=NW)
        self.show_animal_name_dropdown.grid(row=3, column=0, sticky=NW)
        self.outside_dropdown.grid(row=4, column=0, sticky=NW)
        self.core_cnt_dropdown.grid(row=5, column=0, sticky=NW)
        #self.gpu_dropdown.grid(row=5, column=0, sticky=NW)

        self.body_parts_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT NUMBER OF ANIMAL(S)", icon_name='pose', icon_link=Links.ROI_DATA_PLOT.value)
        self.animal_cnt_dropdown = SimBADropDown(parent=self.body_parts_frm, dropdown_options=list(range(1, self.animal_cnt + 1)), dropdown_width=self.longest_animal_name_len, label="NUMBER OF ANIMALS:", label_width=35, value=1, command=self.__create_animal_bp_table)
        self.__create_animal_bp_table(bp_cnt=1)
        self.body_parts_frm.grid(row=1, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)

        self.single_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SINGLE VIDEO", icon_name='video', icon_link=Links.ROI_DATA_PLOT.value)
        self.single_video_dropdown = DropDownMenu(self.single_video_frm, "SELECT VIDEO", self.videos_with_rois_and_data, "15", com=lambda x: self.update_file_select_box_from_dropdown(filename=x, fileselectbox=self.select_video_file_select))
        self.select_video_file_select = FileSelect(self.single_video_frm, "", lblwidth="1", file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)], dropdown=self.single_video_dropdown, initialdir=self.video_dir)
        self.select_video_file_select.filePath.set(self.videos_with_rois_and_data[0])
        self.single_video_dropdown.setChoices(self.videos_with_rois_and_data[0])
        self.single_video_btn = SimbaButton(parent=self.single_video_frm, txt="CREATE SINGLE ROI VIDEO", img='rocket', txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.run, cmd_kwargs={'multiple': False})
        self.single_video_frm.grid(row=3, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=0, sticky=NW)
        self.select_video_file_select.grid(row=0, column=1, sticky=NW)
        self.single_video_btn.grid(row=1, column=0, sticky=NW)


        self.all_videos_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="ALL VIDEOS", icon_name='video', icon_link=Links.ROI_DATA_PLOT.value)
        self.all_videos_btn = SimbaButton(parent=self.all_videos_frm, txt=f"CREATE ALL ROI VIDEOS ({len(self.videos_with_rois_and_data)} VIDEO(S) WITH DATA FOUND)", img='rocket', txt_clr='green', font=Formats.FONT_REGULAR.value, cmd=self.run, cmd_kwargs={'multiple': True})
        self.all_videos_frm.grid(row=4, column=0, sticky=NW)
        self.all_videos_btn.grid(row=0, column=0, sticky=NW)
        self.main_frm.mainloop()

    def __create_animal_bp_table(self, bp_cnt: int):
        bp_cnt = int(bp_cnt)
        if hasattr(self, "animal_table_frm"):
            self.animal_table_frm.destroy()
            for k, v in self.bp_dropdown_dict.items(): v.destroy()
            for k, v in self.bp_clr_dict.items(): v.destroy()
            for k, v in self.bp_size_dict.items(): v.destroy()

        self.animal_table_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT BODY-PARTS", icon_name='pose', icon_link=Links.ROI_DATA_PLOT.value)
        animal_name_heading = SimBALabel(parent=self.animal_table_frm, txt='ANIMAL NAME', width=self.longest_animal_name_len, font=Formats.FONT_HEADER.value)
        bp_heading = SimBALabel(parent=self.animal_table_frm, txt='BODY-PART NAME', width=self.longest_animal_name_len + 5, font=Formats.FONT_HEADER.value)
        bp_clr_heading = SimBALabel(parent=self.animal_table_frm, txt='COLOR', width=self.longest_animal_name_len, font=Formats.FONT_HEADER.value)
        bp_size_heading = SimBALabel(parent=self.animal_table_frm, txt='KEY-POINT SIZE', width=15, font=Formats.FONT_HEADER.value)
        animal_name_heading.grid(row=1, column=0, sticky=NW)
        bp_heading.grid(row=1, column=1, sticky=NW)
        bp_clr_heading.grid(row=1, column=2, sticky=NW)
        bp_size_heading.grid(row=1, column=3, sticky=NW)

        self.bp_dropdown_dict, self.bp_clr_dict, self.bp_size_dict = {}, {}, {}
        for cnt in range(bp_cnt):
            animal_name_lbl = SimBALabel(parent=self.animal_table_frm, txt=self.multi_animal_id_list[cnt], width=self.longest_animal_name_len, font=Formats.FONT_HEADER.value)
            self.bp_dropdown_dict[cnt] = SimBADropDown(parent=self.animal_table_frm, dropdown_options=self.body_parts_lst, dropdown_width=self.longest_animal_name_len, label=None, label_width=None, value=self.body_parts_lst[cnt])
            self.bp_clr_dict[cnt] = SimBADropDown(parent=self.animal_table_frm, dropdown_options=list(self.clr_dict.keys()), dropdown_width=self.longest_animal_name_len + 5, label=None, label_width=None, value=list(self.clr_dict.keys())[cnt])
            self.bp_size_dict[cnt] = SimBADropDown(parent=self.animal_table_frm, dropdown_options=BP_SIZE_OPTIONS, dropdown_width=15, label=None, label_width=None, value='AUTO')
            animal_name_lbl.grid(row=cnt + 2, column=0, sticky=NW)
            self.bp_dropdown_dict[cnt].grid(row=cnt + 2, column=1, sticky=NW)
            self.bp_clr_dict[cnt].grid(row=cnt + 2, column=2, sticky=NW)
            self.bp_size_dict[cnt].grid(row=cnt + 2, column=3, sticky=NW)

        self.animal_table_frm.grid(row=2, column=0, sticky=NW)

    def _disable_clr(self, setting):
        for k, v in self.bp_clr_dict.items():
            if not str_2_bool(setting): v.disable()
            else: v.enable()
        for k, v in self.bp_size_dict.items():
            if not str_2_bool(setting): v.disable()
            else: v.enable()

    def run(self, multiple: bool):
        if multiple:
            video_paths = [v for k, v in self.video_file_paths.items() if k in self.videos_with_rois_and_data]
        else:
            video_name = self.single_video_dropdown.getChoices().rsplit(".", 1)[0]
            video_paths = [self.video_file_paths[video_name]]
        for video_path in video_paths:
            self.check_if_selected_video_path_exist_in_project(video_path=video_path)
        check_float(name="Body-part probability threshold", value=self.threshold_entry_box.entry_get, min_value=0.0, max_value=1.0)
        show_pose = str_2_bool(self.show_pose_estimation_dropdown.get_value())
        show_names = str_2_bool(self.show_animal_name_dropdown.get_value())
        gpu = str_2_bool(self.gpu_dropdown.get_value())
        core_cnt = int(self.core_cnt_dropdown.get_value())
        outside_roi = str_2_bool(self.outside_dropdown.get_value())
        style_attr = {"show_body_part": show_pose, "show_animal_name": show_names}
        body_parts = [v.get_value() for k, v in self.bp_dropdown_dict.items()]
        bp_clrs, bp_sizes = None, None
        if show_pose:
            bp_clrs = [self.clr_dict[v.get_value()] for k, v in self.bp_clr_dict.items()]
            bp_sizes = [v.get_value() for k, v in self.bp_size_dict.items()]

        for video_path in video_paths:
            if core_cnt == 1:
                roi_plotter = ROIPlotter(config_path=self.config_path,
                                         video_path=video_path,
                                         style_attr=style_attr,
                                         threshold=float(self.threshold_entry_box.entry_get),
                                         body_parts=body_parts,
                                         bp_colors=bp_clrs,
                                         bp_sizes=bp_sizes,
                                         outside_roi=outside_roi)
                threading.Thread(target=roi_plotter.run()).start()

            else:
                roi_plotter = ROIPlotMultiprocess(config_path=self.config_path,
                                                  video_path=video_path,
                                                  style_attr=style_attr,
                                                  threshold=float(self.threshold_entry_box.entry_get),
                                                  body_parts=body_parts,
                                                  bp_colors=bp_clrs,
                                                  bp_sizes=bp_sizes,
                                                  core_cnt=core_cnt,
                                                  gpu=gpu,
                                                  outside_roi=outside_roi)
                roi_plotter.run()
                time.sleep(3)
        #


#_ = VisualizeROITrackingPopUp(config_path=r"C:\troubleshooting\roi_entries\project_folder\project_config.ini")


#_ = VisualizeROITrackingPopUp(config_path=r"C:\troubleshooting\roi_duplicates\project_folder\project_config.ini")


# _ = VisualizeROITrackingPopUp(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")
#_ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini')
# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini')
# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/mouse_open_field/project_folder/project_config.ini')
# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/dorian_2/project_folder/project_config.ini')
